from copy import deepcopy
from typing import Dict
from types import SimpleNamespace

from sklearn.preprocessing import OrdinalEncoder

from lightwood.analysis.base import BaseAnalysisBlock


class ConfStats(BaseAnalysisBlock):
    """
    Computes confidence-related statistics on the held-out validation dataset.

    TODO: regression & forecasting tasks
    """

    def __init__(self, deps=('ICP',), ece_bins: int = 10):
        super().__init__(deps=deps)
        self.ece_bins = ece_bins
        self.ordenc = OrdinalEncoder()
        self.n_decimals = 3

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        ns = SimpleNamespace(**kwargs)

        if ns.is_classification:
            possible_labels = ns.stats_info.train_observed_classes
            self.ordenc.fit([[label] for label in possible_labels])
            task_type = 'categorical'
        elif ns.is_multi_ts:
            task_type = 'multi_ts'
        elif ns.is_numerical:
            task_type = 'numerical'
        else:
            return info

        ces, ece, mce, gscore = self._get_stats(info['result_df'],
                                                ns.normal_predictions,
                                                ns.data,
                                                ns.target,
                                                task_type)
        info['maximum_calibration_error'] = round(mce, self.n_decimals)
        info['expected_calibration_error'] = round(ece, self.n_decimals)
        info['binned_conf_acc_difference'] = ces
        info['global_calibration_score'] = round(gscore, self.n_decimals)
        return info

    def _get_stats(self, confs, preds, data, target, task_type='categorical'):
        """
        Computes expected and maximum calibration error for classification tasks.

        Amount of bins is specified by `self.ece_bins`. Data is sorted by increasing confidence prior to binning.

        :return:
        bins: bin-wise absolute difference between estimated confidence and true accuracy.
        ece: weighted average of all bins.
        mce: maximum value in `bins`.
        global_score: 1.0 minus absolute difference between accuracy and confidence over the entire validation set.
        """
        confs = deepcopy(confs).reset_index(drop=True)
        sorted_preds = deepcopy(preds).reset_index(drop=True)
        sorted_inp = deepcopy(data).reset_index(drop=True)
        sorted_val = confs.sort_values(by='confidence', kind='stable')
        sorted_inp['__mdb_confidence'] = sorted_val['confidence']

        if task_type == 'categorical':
            sorted_inp['__mdb_prediction'] = sorted_preds['prediction']
        else:
            if isinstance(confs['lower'][0], list):
                sorted_inp['__mdb_lower'] = confs['lower'].apply(lambda x: x[0])
                sorted_inp['__mdb_upper'] = confs['upper'].apply(lambda x: x[0])
            else:
                sorted_inp['__mdb_lower'] = confs['lower']
                sorted_inp['__mdb_upper'] = confs['upper']

            sorted_inp['__mdb_hits'] = (sorted_inp['__mdb_lower'] <= sorted_inp[target]) & \
                                       (sorted_inp[target] <= sorted_inp['__mdb_upper'])

        size = round(len(sorted_inp) / self.ece_bins)
        bins = []
        ece = 0

        for i in range(1, self.ece_bins):
            bin = sorted_inp.iloc[(i - 1) * size:i * size]

            if len(bin) > 0:
                if task_type == 'categorical':
                    acc = sum(bin[target] == bin['__mdb_prediction']) / size
                else:
                    acc = sum(bin['__mdb_hits'].astype(int)) / len(bin)

                conf = sum(bin['__mdb_confidence']) / size
                gap = abs(acc - conf)
                bins.append(gap)
                ece += gap

        ece /= self.ece_bins
        mce = max(bins) if bins else 0

        if task_type == 'categorical':
            global_acc = sum(sorted_inp[target] == sorted_inp['__mdb_prediction']) / len(sorted_inp)
        else:
            global_acc = sum(sorted_inp['__mdb_hits'].astype(int)) / len(sorted_inp)
        global_conf = sorted_inp['__mdb_confidence'].mean()
        global_score = 1.0 - abs(global_acc - global_conf)

        return bins, ece, mce, global_score



import os
import logging
logging.getLogger('matplotlib').setLevel(level=logging.WARNING)
from lightwood.api import __all__ as api_all_list
from lightwood.api import * # noqa
from lightwood import data
from lightwood.__about__ import __package_name__ as name, __version__


os.environ['TOKENIZERS_PARALLELISM'] = 'true'
__all__ = ['data', 'name', '__version__', *api_all_list]



from lightwood.encoder.categorical.onehot import OneHotEncoder
from lightwood.encoder.categorical.simple_label import SimpleLabelEncoder
from lightwood.encoder.categorical.multihot import MultiHotEncoder
from lightwood.encoder.categorical.autoencoder import CategoricalAutoEncoder

__all__ = ['OneHotEncoder', 'SimpleLabelEncoder', 'MultiHotEncoder', 'CategoricalAutoEncoder']



import torch
import numpy as np
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder, OrdinalEncoder


class MinMaxNormalizer:
    def __init__(self, combination=()):
        self.scaler = MinMaxScaler()
        self.abs_mean = None
        self.combination = combination  # tuple with values in grouped-by columns
        self.output_size = 1

    def prepare(self, x: np.ndarray) -> None:
        # @TODO: streamline input type
        if isinstance(x[0], list):
            x = np.vstack(x)
        if isinstance(x[0], torch.Tensor):
            x = torch.stack(x).numpy()
        if len(x.shape) < 2:
            x = np.expand_dims(x, axis=1)

        x[x == None] = 0 # noqa
        x = x.astype(float)
        self.abs_mean = np.mean(np.abs(x))
        self.scaler.fit(x.reshape(x.size, -1))

    def encode(self, y: np.ndarray) -> torch.Tensor:
        if isinstance(y[0], list):
            y = np.vstack(y)
        if isinstance(y[0], torch.Tensor):
            y = torch.stack(y).numpy()
        if len(y.shape) < 2:
            y = np.expand_dims(y, axis=1)

        shape = y.shape
        y = y.astype(float).reshape(-1, self.scaler.n_features_in_)
        out = torch.reshape(torch.Tensor(self.scaler.transform(y)), shape)
        return out

    def decode(self, y):
        return self.scaler.inverse_transform(y)


class CatNormalizer:
    def __init__(self, encoder_class='one_hot'):
        self.encoder_class = encoder_class
        if encoder_class == 'one_hot':
            self.scaler = OneHotEncoder(sparse=False, handle_unknown='ignore')
        else:
            self.scaler = OrdinalEncoder()

        self.unk = "<UNK>"

    def prepare(self, x):
        X = set()
        for i in x:
            for j in i:
                X.add(j if j is not None else self.unk)
        self.scaler.fit(np.array(list(X)).reshape(-1, 1))
        self.output_size = len(self.scaler.categories_[0]) if self.encoder_class == 'one_hot' else 1

    def encode(self, Y):
        y = [[str(j) if j is not None else self.unk for j in i] for i in Y]
        y = [[j if j in self.scaler.categories_[0] else self.unk for j in i] for i in y]
        y = np.array(y)
        out = []
        for i in y:
            transformed = self.scaler.transform(i.reshape(-1, 1))
            if isinstance(self.scaler, OrdinalEncoder):
                transformed = transformed.flatten()
            out.append(transformed)

        return torch.Tensor(out)

    def decode(self, y):
        return self.scaler.inverse_transform(y)



from typing import Iterable
from type_infer.helpers import is_nan_numeric


def filter_nan_and_none(series: Iterable) -> list:
    return [x for x in series if not is_nan_numeric(x) and x is not None]



from typing import Dict, Optional

import torch
from tab_transformer_pytorch import TabTransformer

from lightwood.helpers.device import get_device_from_name
from lightwood.data.encoded_ds import EncodedDs
from lightwood.encoder.base import BaseEncoder
from lightwood.mixer.neural import Neural


class TabTransformerMixer(Neural):
    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            target_encoder: BaseEncoder,
            fit_on_dev: bool,
            search_hyperparameters: bool,
            train_args: Optional[dict] = None
    ):
        """
        This mixer trains a TabTransformer network (FT variant), using concatenated encoder outputs for each dataset feature as input, to predict the encoded target column representation as output.
        
        Training logic is based on the Neural mixer, please refer to it for more details on each input parameter.
        """  # noqa
        self.train_args = train_args if train_args else {}
        super().__init__(
            stop_after,
            target,
            dtype_dict,
            target_encoder,
            'FTTransformer',
            False,  # fit_on_dev
            search_hyperparameters,
            n_epochs=self.train_args.get('n_epochs', None)
        )
        self.lr = self.train_args.get('lr')
        self.stable = False  # still experimental

    def _init_net(self, ds: EncodedDs):
        self.net_class = TabTransformer
        self.model = TabTransformer(
            categories=(),                                                      # unused, everything is numerical by now
            num_continuous=len(ds[0][0]),
            dim=self.train_args.get('dim', 32),
            dim_out=self.train_args.get('dim_out', len(ds[0][1])),
            depth=self.train_args.get('depth', 6),
            heads=self.train_args.get('heads', 8),
            attn_dropout=self.train_args.get('attn_dropout', 0.1),              # post-attention dropout
            ff_dropout=self.train_args.get('ff_dropout', 0.1),                  # feed forward dropout
            mlp_hidden_mults=self.train_args.get('mlp_hidden_mults', (4, 2)),   # relative multiples of each hidden dimension of the last mlp to logits  # noqa
            # mlp_act=self.train_args.get('mlp_act', nn.ReLU()),                # TODO: import str from nn activations
        )
        self.model.device = get_device_from_name('')
        self.model.to(self.model.device)

    def _net_call(self, x: torch.Tensor) -> torch.Tensor:
        x = torch.unsqueeze(x, 0) if len(x.shape) < 2 else x
        return self.model(torch.Tensor(), x)

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """ Skip the usual partial_fit call at the end. """  # noqa
        self._fit(train_data, dev_data)



import torch
from lightwood.helpers.torch import LightwoodAutocast


# Basically cross entropy loss that does the one hot decoding of the targets inside of it... useful for code-logic reasons to have it setup like this # noqa
class TransformCrossEntropyLoss(torch.nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        self.cross_entropy_loss = torch.nn.CrossEntropyLoss(**kwargs)

    def forward(self, preds, target):
        with LightwoodAutocast():
            cat_labels = target.max(1).indices
            return self.cross_entropy_loss(preds, cat_labels)



import importlib
from copy import deepcopy
from datetime import datetime
from typing import Dict, Union, Optional

import optuna
import numpy as np
import pandas as pd
from sktime.forecasting.compose import TransformedTargetForecaster
from sktime.forecasting.base import ForecastingHorizon, BaseForecaster
from sktime.performance_metrics.forecasting import MeanAbsolutePercentageError
from sktime.forecasting.statsforecast import StatsForecastAutoARIMA as AutoARIMA

from lightwood.helpers.log import log
from lightwood.helpers.templating import _add_cls_kwarg
from lightwood.mixer.base import BaseMixer
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class SkTime(BaseMixer):
    forecaster: str
    horizon: int
    target: str
    supports_proba: bool
    model_path: str
    hyperparam_search: bool

    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            horizon: int,
            ts_analysis: Dict,
            model_path: str = None,
            model_kwargs: Optional[dict] = None,
            auto_size: bool = True,
            sp: int = None,
            hyperparam_search: bool = True,
            use_stl: bool = False
    ):
        """
        This mixer is a wrapper around the popular time series library sktime. It exhibits different behavior compared
        to other forecasting mixers, as it predicts based on indices in a forecasting horizon that is defined with
        respect to the last seen data point at training time.
        
        Due to this, the mixer tries to "fit_on_all" so that the latest point in the validation split marks the 
        difference between training data and where forecasts will start. In practice, you need to specify how much 
        time has passed since the aforementioned timestamp for correct forecasts. By default, it is assumed that
         predictions are for the very next timestamp post-training.
        
        If the task has groups (i.e. 'TimeseriesSettings.group_by' is not empty), the mixer will spawn one forecaster 
        object per each different group observed at training time, plus an additional default forecaster fit with all data.
        
        There is an optuna-based automatic hyperparameter search. For now, it considers selecting the forecaster type
        based on the global SMAPE error across all groups.
        
        :param stop_after: time budget in seconds.
        :param target: column to forecast.
        :param dtype_dict: dtypes of all columns in the data.
        :param horizon: length of forecasted horizon.
        :param sp: seasonality period to enforce (instead of automatic inference done at the `ts_analysis` module)
        :param ts_analysis: dictionary with miscellaneous time series info, as generated by 'lightwood.data.timeseries_analyzer'.
        :param model_path: sktime forecaster to use as underlying model(s). Should be a string with format "$module.$class' where '$module' is inside `sktime.forecasting`. Default is 'arima.AutoARIMA'.
        :param model_kwargs: specifies additional paramters to pass to the model if model_path is provided. 
        :param hyperparam_search: bool that indicates whether to perform the hyperparameter tuning or not.
        :param auto_size: whether to filter out old data points if training split is bigger than a certain threshold (defined by the dataset sampling frequency). Enabled by default to avoid long training times in big datasets.
        :param use_stl: Whether to use de-trenders and de-seasonalizers fitted in the timeseries analysis phase.
        """  # noqa
        super().__init__(stop_after)

        assert ts_analysis['tss'].horizon > 1, log.error("Horizon must be greater than 1 when using the SkTime Mixer!")

        self.stable = False
        self.prepared = False
        self.supports_proba = False
        self.target = target
        self.name = 'AutoSKTime'

        default_possible_models = [
            'croston.Croston',
            'theta.ThetaForecaster',
            'trend.STLForecaster',
            'trend.PolynomialTrendForecaster',
        ]

        self.dtype_dict = dtype_dict
        self.ts_analysis = ts_analysis
        self.horizon = horizon
        self.sp = sp
        self.grouped_by = ['__default'] if not ts_analysis['tss'].group_by else ts_analysis['tss'].group_by
        self.auto_size = auto_size
        self.cutoff_factor = 4  # times the detected maximum seasonal period
        self.use_stl = use_stl

        # optuna hyperparameter tuning
        self.models = {}
        self.cutoffs = {}  # last seen timestamp per each model
        self.study = None
        self.hyperparam_dict = {}
        self.model_path = model_path if model_path else 'trend.STLForecaster'
        self.model_kwargs = model_kwargs if model_kwargs else {}
        self.hyperparam_search = hyperparam_search
        self.trial_error_fn = MeanAbsolutePercentageError(symmetric=True)
        self.possible_models = default_possible_models if not model_path else [model_path]
        self.n_trials = len(self.possible_models)
        self.freq = self._get_freq(self.ts_analysis['deltas']['__default'])

        # sktime forecast horizon object is made relative to the end of the latest data point seen at training time
        # the default assumption is to forecast the next `self.horizon` after said data point
        self.fh = ForecastingHorizon(np.arange(1, self.horizon + 1), is_relative=True)

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits a set of sktime forecasters. The number of models depends on how many groups are observed at training time.

        Forecaster type can be specified by providing the `model_class` argument in `__init__()`. It can also be determined by hyperparameter optimization based on dev data validation error.
        """  # noqa
        log.info(f'Started fitting {self.name} forecaster for array prediction')

        if self.hyperparam_search:
            search_space = {'class': self.possible_models}
            self.study = optuna.create_study(direction='minimize', sampler=optuna.samplers.GridSampler(search_space))
            self.study.optimize(lambda trial: self._get_best_model(trial, train_data, dev_data), n_trials=self.n_trials)
            data = ConcatedEncodedDs([train_data, dev_data])
            self._fit(data)

        else:
            data = ConcatedEncodedDs([train_data, dev_data])
            self._fit(data)

    def _fit(self, data):
        """
        Internal method that fits forecasters to a given dataframe.
        """
        df = data.data_frame.sort_values(by=f'__mdb_original_{self.ts_analysis["tss"].order_by}')
        gby = self.ts_analysis['tss'].group_by

        if not self.hyperparam_search and not self.study:
            module_name = self.model_path
        else:
            finished_study = sum([int(trial.state.is_finished()) for trial in self.study.trials]) == self.n_trials
            if finished_study:
                log.info(f'Selected best model: {self.study.best_params["class"]}')
                module_name = self.study.best_params['class']
            else:
                module_name = self.hyperparam_dict['class']  # select active optuna choice

        sktime_module = importlib.import_module('.'.join(['sktime', 'forecasting', module_name.split(".")[0]]))
        try:
            model_class = getattr(sktime_module, module_name.split(".")[1])
        except AttributeError:
            model_class = AutoARIMA  # use AutoARIMA when the provided class does not exist

        grouped = df.groupby(by=gby) if gby else df.groupby(lambda x: '__default')
        for group, series_data in grouped:
            kwargs = {}
            sp = self.sp if self.sp else self.ts_analysis['periods'].get(group, [1])[0]

            options = self.model_kwargs
            options['sp'] = sp               # seasonality period
            options['suppress_warnings'] = True  # ignore warnings if possible
            options['error_action'] = 'raise'    # avoids fit() failing silently

            if self.model_path == 'fbprophet.Prophet':
                options['freq'] = self.freq

            for k, v in options.items():
                kwargs = _add_cls_kwarg(model_class, kwargs, k, v)

            model_pipeline = []

            if self.use_stl and self.ts_analysis['stl_transforms'].get(group, False):
                model_pipeline.insert(0, ("detrender",
                                          self.ts_analysis['stl_transforms'][group]["transformer"].detrender))
                model_pipeline.insert(0, ("deseasonalizer",
                                          self.ts_analysis['stl_transforms'][group]["transformer"].deseasonalizer))
                kwargs['sp'] = None

            model_pipeline.append(("forecaster", model_class(**kwargs)))

            self.models[group] = TransformedTargetForecaster(model_pipeline)

            oby_col = self.ts_analysis['tss'].order_by
            if self.grouped_by == ['__default']:
                series_oby = df[oby_col]
            else:
                series_oby = series_data[oby_col]

            self.cutoffs[group] = series_oby.index[-1]  # defines the 'present' time for each partition
            series = series_data[self.target]
            if series_data.size > self.ts_analysis['tss'].window:
                series = series.sort_index(ascending=True)
                series = series.reset_index(drop=True)
                series = series.loc[~pd.isnull(series.values)]  # remove NaN  # @TODO: benchmark imputation vs this?

                if self.model_path == 'fbprophet.Prophet':
                    try:
                        series = self._transform_index_to_datetime(series, series_oby, options['freq'])
                    except Exception:
                        if group == '__default':
                            # out of bounds with true freq in __default group is fine, we skip it
                            continue

                series = series.astype(float)

                # if data is huge, filter out old records for quicker fitting
                if self.auto_size:
                    cutoff = min(len(series), max(500, options.get('sp', 1) * self.cutoff_factor))
                    series = series.iloc[-cutoff:]
                try:
                    self.models[group].fit(series, fh=self.fh)
                except Exception:
                    self.models[group] = model_class()  # with default options (i.e. no seasonality, among others)
                    self.models[group].fit(series, fh=self.fh)

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        """
        Note: sktime asks for "specification of the time points for which forecasts are requested", and this mixer complies by assuming forecasts will start immediately after the last observed value.

        Because of this, `ProblemDefinition.fit_on_all` is set to True so that `partial_fit` uses both `dev` and `test` splits to fit the models.

        Due to how lightwood implements the `update` procedure, expected inputs for this method are:

        :param dev_data: original `test` split (used to validate and select model if ensemble is `BestOf`).
        :param train_data: concatenated original `train` and `dev` splits.
        """  # noqa
        self.hyperparam_search = False
        self.fit(dev_data, train_data)
        self.prepared = True

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Calls the mixer to emit forecasts.
        
        If there are groups that were not observed at training, a default forecaster (trained on all available data) is used, warning the user that performance might not be optimal.
        
        Latest data point in `train_data` passed to `fit()` determines the starting point for predictions. Relative offsets will be automatically determined when predicting for other starting points.
        """  # noqa
        if args.predict_proba:
            log.warning('This mixer does not output probability estimates')

        df = deepcopy(ds.data_frame)
        df = df.rename_axis('__sktime_index').reset_index()

        gby = self.ts_analysis['tss'].group_by
        ydf = pd.DataFrame(0,  # zero-filled
                           index=df.index,
                           columns=['prediction'],
                           dtype=object)

        pending_idxs = set(df.index)
        grouped = df.groupby(by=gby) if gby else df.groupby(lambda x: '__default')
        for group, series_data in grouped:
            if series_data.size > 0:
                start_ts = series_data['__sktime_index'].iloc[0]
                series = series_data[self.target]
                series_idxs = series.index
                if self.models.get(group, False) and self.models[group].is_fitted:
                    freq = self.ts_analysis['deltas'][group]
                    delta = (start_ts - self.cutoffs[group]).total_seconds()
                    offset = round(delta / freq)
                    forecaster = self.models[group]
                    ydf = self._call_groupmodel(ydf, forecaster, series, offset=offset)
                else:
                    log.warning(f"Applying naive forecaster for novel group {group}. Performance might not be optimal.")
                    ydf = self._call_default(ydf, series.values, series_idxs)
                pending_idxs -= set(series_idxs)

        # apply default model in all remaining novel-group rows
        if len(pending_idxs) > 0:
            series = df[self.target][list(pending_idxs)].squeeze()
            ydf = self._call_default(ydf, series, list(pending_idxs))

        return ydf[['prediction']]

    def _call_groupmodel(self,
                         ydf: pd.DataFrame,
                         model: BaseForecaster,
                         series: pd.Series,
                         offset: int = 0):
        """
        Inner method that calls a `sktime.BaseForecaster`.

        :param offset: indicates relative offset to the latest data point seen during model training. Cannot be less than the number of training data points + the amount of diffences applied internally by the model.
        """  # noqa
        if isinstance(model, TransformedTargetForecaster):
            submodel = model.steps_[-1][-1]
        else:
            submodel = model

        min_offset = -len(submodel._y) + 1
        if hasattr(submodel, 'd'):
            model_d = 0 if submodel.d is None else submodel.d
            min_offset += model_d

        start = max(offset, min_offset)
        end = start + series.shape[0] + self.horizon

        # Workaround for StatsForecastAutoARIMA (see sktime#3600)
        if isinstance(submodel, AutoARIMA):
            all_preds = model.predict(np.arange(min_offset, end)).tolist()[-(end - start):]
        else:
            all_preds = model.predict(np.arange(start, end)).tolist()

        for true_idx, (idx, _) in enumerate(series.items()):
            start_idx = max(0, true_idx)
            end_idx = start_idx + self.horizon
            ydf['prediction'].loc[idx] = all_preds[start_idx:end_idx]
        return ydf

    def _call_default(self, ydf, data, idxs):
        # last value from each window equals shifted target (by 1)  # noqa
        series = np.array([0] + list(data.flatten())[:-1])
        all_preds = [[value for _ in range(self.horizon)] for value in series]
        ydf['prediction'].iloc[idxs] = all_preds
        return ydf

    def _get_best_model(self, trial, train_data, test_data):
        """
        Helper function for Optuna hyperparameter optimization.
        For now, it uses dev data split to find the best model out of the list specified in self.possible_models.
        """

        self.hyperparam_dict = {
            'class': trial.suggest_categorical('class', self.possible_models)
        }
        log.info(f'Starting trial with hyperparameters: {self.hyperparam_dict}')
        try:
            self._fit(train_data)
            y_true = test_data.data_frame[self.target].values[:self.horizon]
            y_pred = pd.DataFrame(self(test_data)['prediction'].iloc[0][:len(y_true)])
            error = self.trial_error_fn(y_true, y_pred)
        except Exception as e:
            log.debug(e)
            error = np.inf

        log.info(f'Trial got error: {error}')
        return error

    def _transform_index_to_datetime(self, series, series_oby, freq):
        series_oby = np.array([np.array(lst) for lst in series_oby])
        start = datetime.utcfromtimestamp(np.min(series_oby[series_oby != np.min(series_oby)]))
        series.index = pd.date_range(start=start, freq=freq, normalize=False, periods=series.shape[0])
        return series

    def _get_freq(self, delta):
        labels = ['S', 'T', 'H', 'D', 'W', 'M', 'Q', 'Y']
        secs = [1,
                60,
                60 * 60,
                60 * 60 * 24,
                60 * 60 * 24 * 7,
                60 * 60 * 24 * 7 * 4,
                60 * 60 * 24 * 7 * 4 * 3,
                60 * 60 * 24 * 7 * 4 * 12]
        min_diff = np.argmin(np.abs(np.array(secs) - delta))
        return labels[min_diff]



from copy import deepcopy
from types import SimpleNamespace
from typing import Dict, Optional, Tuple

import pandas as pd
import numpy as np
from sklearn.preprocessing import OrdinalEncoder

from pyod.models.lof import LOF
from pyod.models.pca import PCA
from pyod.models.hbos import HBOS
from pyod.models.suod import SUOD

from lightwood.analysis.base import BaseAnalysisBlock
from lightwood.helpers.log import log
from lightwood.helpers.parallelism import get_nr_procs


class PyOD(BaseAnalysisBlock):
    """
    Wrapper analysis block for the 'PyOD' anomaly detection library.

    For now, the following techniques are supported:
      - SUOD: Large-scale unsupervised heterogeneous outlier detection
    """  # noqa

    def __init__(self, contamination=0.1, deps: Optional[Tuple] = ...):
        super().__init__(deps=deps)
        self.target = None
        self.input_cols = []
        self.exceptions = ['__make_predictions']
        self.ordinal_encoders = dict()
        if contamination > 0.5:
            log.warning("Contamination higher than maximum possible value, setting at 0.5...")
        self.contamination = min(contamination, 0.5)

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        log.info('Preparing to compute anomaly detection with PyOD')
        ns = SimpleNamespace(**kwargs)
        self.target = ns.target
        detector_list = [LOF(n_neighbors=10, contamination=self.contamination),
                         LOF(n_neighbors=30, contamination=self.contamination),
                         HBOS(contamination=self.contamination),
                         PCA(contamination=self.contamination)]

        df = deepcopy(ns.train_data.data_frame)
        n_jobs = get_nr_procs(df)
        clf = SUOD(base_estimators=detector_list,
                   contamination=self.contamination,
                   n_jobs=n_jobs,
                   combination='average',
                   verbose=False)

        if ns.tss.is_timeseries:
            for gcol in ns.tss.group_by:
                self.ordinal_encoders[gcol] = OrdinalEncoder(handle_unknown='use_encoded_value', unknown_value=-1)
                self.ordinal_encoders[gcol].fit(df[gcol].values.reshape(-1, 1))
            for i in range(1, ns.tss.horizon):
                self.exceptions.append(f'{self.target}_timestep_{i}')

            self.exceptions.append(ns.tss.order_by)
            df = self._preprocess_ts_df(df, ns)

        for col in df.columns:
            if col != self.target and '__mdb' not in col and col not in self.exceptions:
                self.input_cols.append(col)

        df = df.loc[:, self.input_cols].dropna()
        clf.fit(df.values)
        info['pyod_explainer'] = clf
        return info

    def explain(self,
                row_insights: pd.DataFrame,
                global_insights: Dict[str, object],
                **kwargs
                ) -> Tuple[pd.DataFrame, Dict[str, object]]:
        log.info('Performing anomaly detection with PyOD...')
        ns = SimpleNamespace(**kwargs)
        pyod_explainer = ns.analysis.get('pyod_explainer', None)
        if pyod_explainer is None:
            return row_insights, global_insights

        df = deepcopy(ns.data)
        if ns.tss.is_timeseries:
            df = self._preprocess_ts_df(df, ns)

        df = df[self.input_cols].fillna(0)
        row_insights['pyod_anomaly'] = pyod_explainer.predict(df).astype(bool)  # binary labels (0: inlier, 1: outlier)
        return row_insights, global_insights

    def _preprocess_ts_df(self, df: pd.DataFrame, ns: SimpleNamespace) -> pd.DataFrame:
        for gcol in ns.tss.group_by:
            df[gcol] = self.ordinal_encoders[gcol].transform(df[gcol].values.reshape(-1, 1).astype(np.object)).flatten()

        for w in range(ns.tss.window + 1):
            df[f'__pyod_window_{ns.tss.window - w}'] = df[f'__mdb_ts_previous_{self.target}'].apply(lambda x: x[w])
        return df



from lightwood.helpers.device import is_cuda_compatible, get_devices
from lightwood.helpers.general import is_none
from lightwood.helpers.ts import get_ts_groups, get_inferred_timestamps, add_tn_num_conf_bounds, add_tn_cat_conf_bounds
from lightwood.helpers.io import read_from_path_or_url
from lightwood.helpers.parallelism import get_nr_procs, mut_method_call, run_mut_method
from lightwood.helpers.numeric import filter_nan_and_none
from lightwood.helpers.seed import seed
from lightwood.helpers.torch import average_vectors, concat_vectors_and_pad, LightwoodAutocast


__all__ = ['is_cuda_compatible', 'get_devices', 'mut_method_call', 'run_mut_method',
           'get_ts_groups', 'is_none', 'read_from_path_or_url', 'seed',
           'get_inferred_timestamps', 'add_tn_num_conf_bounds', 'add_tn_cat_conf_bounds', 'get_nr_procs',
           'average_vectors', 'concat_vectors_and_pad', 'LightwoodAutocast', 'filter_nan_and_none' ]  # noqa



import os
import sys
import time
import string
import random
import tempfile
import importlib

from copy import deepcopy
from types import ModuleType

from type_infer.dtype import dtype

from lightwood.helpers.log import log
from lightwood.api.types import JsonAI
from lightwood.api.json_ai import add_implicit_values, lookup_encoder
from lightwood.helpers.constants import IMPORTS, IMPORT_EXTERNAL_DIRS
from lightwood.helpers.templating import call, inline_dict, align
from lightwood.__about__ import __version__ as lightwood_version


def code_from_json_ai(json_ai: JsonAI) -> str:
    """
    Generates a custom ``PredictorInterface`` given the specifications from ``JsonAI`` object.

    :param json_ai: ``JsonAI`` object with fully specified parameters

    :returns: Automated syntax of the ``PredictorInterface`` object.
    """
    json_ai = deepcopy(json_ai)
    # ----------------- #
    # Fill in any missing values
    json_ai = add_implicit_values(json_ai)

    # ----------------- #

    # Instantiate data types
    dtype_dict = {}

    for k in json_ai.dtype_dict:
        if json_ai.dtype_dict[k] not in (dtype.invalid, dtype.empty):
            dtype_dict[k] = json_ai.dtype_dict[k]

    # Populate imputers
    imputer_dict = {}
    if json_ai.imputers:
        for imputer in json_ai.imputers:
            imputer_dict[imputer['args']['target'].replace('\'', '').replace('\"', '')] = call(imputer)
    json_ai.imputers = imputer_dict
    imputers = inline_dict(json_ai.imputers)

    # Populate encoders
    encoder_dict = {}
    for col_name, encoder in json_ai.encoders.items():
        encoder_dict[col_name] = call(encoder)

    # Populate time-series specific details
    # TODO: consider moving this to a `JsonAI override` phase
    tss = json_ai.problem_definition.timeseries_settings
    if tss.is_timeseries:
        if tss.use_previous_target:
            col_name = f"__mdb_ts_previous_{json_ai.problem_definition.target}"
            target_type = json_ai.dtype_dict[json_ai.problem_definition.target]
            json_ai.problem_definition.timeseries_settings.target_type = target_type
            encoder_dict[col_name] = call(
                lookup_encoder(
                    target_type,
                    col_name,
                    False,
                    json_ai.problem_definition,
                    False,
                    None,
                )
            )

            dtype_dict[col_name] = target_type
            # @TODO: Is populating the json_ai at this stage even necessary?
            json_ai.encoders[col_name] = encoder_dict[col_name]
            json_ai.dtype_dict[col_name] = target_type
            json_ai.dependency_dict[col_name] = []

    # ----------------- #

    input_cols = [x.replace("'", "\\'").replace('"', '\\"') for x in json_ai.encoders
                  if x != json_ai.problem_definition.target]
    if len(input_cols) < 1:
        raise Exception('There are no valid input features. Please check your data before trying again.')
    input_cols = ",".join([f"""'{name}'""" for name in input_cols])

    # ----------------- #
    # Time-series specific code blocks
    # ----------------- #

    ts_transform_code = ""
    ts_analyze_code = None
    ts_encoder_code = ""
    if json_ai.timeseries_transformer is not None:
        ts_transform_code = f"""
log.info('Transforming timeseries data')
data = {call(json_ai.timeseries_transformer)}
"""
        ts_analyze_code = f"""
self.ts_analysis = {call(json_ai.timeseries_analyzer)}
"""
    # @TODO: set these kwargs/properties in the json ai construction (if possible)
    if json_ai.timeseries_analyzer is not None:
        ts_encoder_code = """
if encoder.is_timeseries_encoder:
    kwargs['ts_analysis'] = self.ts_analysis
"""

    if json_ai.problem_definition.timeseries_settings.is_timeseries:
        ts_target_code = """
if encoder.is_target:
    encoder.normalizers = self.ts_analysis['target_normalizers']
    encoder.group_combinations = self.ts_analysis['group_combinations']
"""
    else:
        ts_target_code = ""

    # ----------------- #
    # Statistical Analysis Body
    # ----------------- #

    analyze_data_body = f"""
self.statistical_analysis = statistical_analysis(data,
                                                 self.dtype_dict,
                                                 self.problem_definition.to_dict(),
                                                 {json_ai.identifiers})

# Instantiate post-training evaluation
self.analysis_blocks = [{', '.join([call(block) for block in json_ai.analysis_blocks])}]
    """

    analyze_data_body = align(analyze_data_body, 2)

    # ----------------- #
    # Pre-processing Body
    # ----------------- #

    clean_body = f"""
log.info('Cleaning the data')
self.imputers = {imputers}
data = {call(json_ai.cleaner)}

# Time-series blocks
{ts_transform_code}
"""

    clean_body += '\nreturn data'

    clean_body = align(clean_body, 2)

    # ----------------- #
    # Train-Test Splitter Body
    # ----------------- #

    split_body = f"""
log.info("Splitting the data into train/test")
train_test_data = {call(json_ai.splitter)}

return train_test_data
    """

    split_body = align(split_body, 2)

    # ----------------- #
    # Prepare features Body
    # ----------------- #

    prepare_body = """
self.mode = 'train'

if self.statistical_analysis is None:
    raise Exception("Please run analyze_data first")
"""
    if ts_analyze_code is not None:
        prepare_body += f"""
if self.mode != 'predict':
    {align(ts_analyze_code, 1)}
"""

    prepare_body += f"""
# Column to encoder mapping
self.encoders = {inline_dict(encoder_dict)}

# Prepare the training + dev data
concatenated_train_dev = pd.concat([data['train'], data['dev']])

prepped_encoders = {{}}

# Prepare input encoders
parallel_encoding = parallel_encoding_check(data['train'], self.encoders)

if parallel_encoding:
    log.debug('Preparing in parallel...')
    for col_name, encoder in self.encoders.items():
        if col_name != self.target and not encoder.is_trainable_encoder:
            prepped_encoders[col_name] = (encoder, concatenated_train_dev[col_name], 'prepare')
    prepped_encoders = mut_method_call(prepped_encoders)

else:
    log.debug('Preparing sequentially...')
    for col_name, encoder in self.encoders.items():
        if col_name != self.target and not encoder.is_trainable_encoder:
            log.debug(f'Preparing encoder for {{col_name}}...')
            encoder.prepare(concatenated_train_dev[col_name])
            prepped_encoders[col_name] = encoder

# Store encoders
for col_name, encoder in prepped_encoders.items():
    self.encoders[col_name] = encoder

# Prepare the target
if self.target not in prepped_encoders:
    if self.encoders[self.target].is_trainable_encoder:
        self.encoders[self.target].prepare(data['train'][self.target], data['dev'][self.target])
    else:
        self.encoders[self.target].prepare(pd.concat([data['train'], data['dev']])[self.target])

# Prepare any non-target encoders that are learned
for col_name, encoder in self.encoders.items():
    if col_name != self.target and encoder.is_trainable_encoder:
        priming_data = pd.concat([data['train'], data['dev']])
        kwargs = {{}}
        if self.dependencies[col_name]:
            kwargs['dependency_data'] = {{}}
            for col in self.dependencies[col_name]:
                kwargs['dependency_data'][col] = {{
                    'original_type': self.dtype_dict[col],
                    'data': priming_data[col]
                }}
            {align(ts_encoder_code, 3)}

        # If an encoder representation requires the target, provide priming data
        if hasattr(encoder, 'uses_target'):
            kwargs['encoded_target_values'] = self.encoders[self.target].encode(priming_data[self.target])

        encoder.prepare(data['train'][col_name], data['dev'][col_name], **kwargs)

    {align(ts_target_code, 1)}
"""
    prepare_body = align(prepare_body, 2)

    # ----------------- #
    # Featurize Data Body
    # ----------------- #

    feature_body = f"""
log.info('Featurizing the data')

tss = self.problem_definition.timeseries_settings

feature_data = dict()
for key, data in split_data.items():
    if key != 'stratified_on':

        # compute and store two splits - full and filtered (useful for time series post-train analysis)
        if key not in self.feature_cache:
            featurized_split = EncodedDs(self.encoders, data, self.target)
            filtered_subset = EncodedDs(self.encoders, filter_ts(data, tss), self.target)

            for k, s in zip((key, f'{{key}}_filtered'), (featurized_split, filtered_subset)):
                self.feature_cache[k] = s

        for k in (key, f'{{key}}_filtered'):
            feature_data[k] = self.feature_cache[k]

return feature_data

"""  # noqa

    feature_body = align(feature_body, 2)

    # ----------------- #
    # Fit Mixer Body
    # ----------------- #

    fit_body = f"""
self.mode = 'train'

# --------------- #
# Extract data
# --------------- #
# Extract the featurized data into train/dev/test
encoded_train_data = enc_data['train']
encoded_dev_data = enc_data['dev']
encoded_test_data = enc_data['test_filtered']

log.info('Training the mixers')

# --------------- #
# Fit Models
# --------------- #
# Assign list of mixers
self.mixers = [{', '.join([call(x) for x in json_ai.model["args"]["submodels"]])}]

# Train mixers
trained_mixers = []
for mixer in self.mixers:
    try:
        if mixer.trains_once:
            self.fit_mixer(mixer,
                           ConcatedEncodedDs([encoded_train_data, encoded_dev_data]),
                           encoded_test_data)
        else:
            self.fit_mixer(mixer, encoded_train_data, encoded_dev_data)
        trained_mixers.append(mixer)
    except Exception as e:
        log.warning(f'Exception: {{e}} when training mixer: {{mixer}}')
        if {json_ai.problem_definition.strict_mode} and mixer.stable:
            raise e

# Update mixers to trained versions
if not trained_mixers:
    raise Exception('No mixers could be trained! Please verify your problem definition or JsonAI model representation.')
self.mixers = trained_mixers

# --------------- #
# Create Ensembles
# --------------- #
log.info('Ensembling the mixer')
# Create an ensemble of mixers to identify best performing model
# Dirty hack
self.ensemble = {call(json_ai.model)}
self.supports_proba = self.ensemble.supports_proba
"""
    fit_body = align(fit_body, 2)

    # ----------------- #
    # Analyze Ensemble Body
    # ----------------- #

    analyze_ensemble = f"""

# --------------- #
# Extract data
# --------------- #
# Extract the featurized data into train/dev/test
encoded_train_data = enc_data['train']
encoded_dev_data = enc_data['dev']
encoded_test_data = enc_data['test']

# --------------- #
# Analyze Ensembles
# --------------- #
log.info('Analyzing the ensemble of mixers')
self.model_analysis, self.runtime_analyzer = {call(json_ai.analyzer)}
"""
    analyze_ensemble = align(analyze_ensemble, 2)

    # ----------------- #
    # Adjust Ensemble Body
    # ----------------- #

    adjust_body = f"""
self.mode = 'train'

# --------------- #
# Prepare data
# --------------- #
if dev_data is None:
    data = train_data
    split = splitter(
        data=data,
        pct_train=0.8,
        pct_dev=0.2,
        pct_test=0,
        tss=self.problem_definition.timeseries_settings.to_dict(),
        seed=self.problem_definition.seed_nr,
        target=self.target,
        dtype_dict=self.dtype_dict)
    train_data = split['train']
    dev_data = split['dev']

if adjust_args is None or not adjust_args.get('learn_call'):
    train_data = self.preprocess(train_data)
    dev_data = self.preprocess(dev_data)

dev_data = EncodedDs(self.encoders, dev_data, self.target)
train_data = EncodedDs(self.encoders, train_data, self.target)

# --------------- #
# Update/Adjust Mixers
# --------------- #
log.info('Updating the mixers')

for mixer in self.mixers:
        mixer.partial_fit(train_data, dev_data, adjust_args)
"""  # noqa

    adjust_body = align(adjust_body, 2)

    # ----------------- #
    # Learn Body
    # ----------------- #

    learn_body = """
self.mode = 'train'
n_phases = 8 if self.problem_definition.fit_on_all else 7

# Perform stats analysis
log.info(f'[Learn phase 1/{n_phases}] - Statistical analysis')
self.analyze_data(data)

# Pre-process the data
log.info(f'[Learn phase 2/{n_phases}] - Data preprocessing')
data = self.preprocess(data)

# Create train/test (dev) split
log.info(f'[Learn phase 3/{n_phases}] - Data splitting')
train_dev_test = self.split(data)

# Prepare encoders
log.info(f'[Learn phase 4/{n_phases}] - Preparing encoders')
self.prepare(train_dev_test)

# Create feature vectors from data
log.info(f'[Learn phase 5/{n_phases}] - Feature generation')
enc_train_test = self.featurize(train_dev_test)

# Prepare mixers
log.info(f'[Learn phase 6/{n_phases}] - Mixer training')
if not self.problem_definition.embedding_only:
    self.fit(enc_train_test)
else:
    self.mixers = []
    self.ensemble = Embedder(self.target, mixers=list(), data=enc_train_test['train'])
    self.supports_proba = self.ensemble.supports_proba

# Analyze the ensemble
log.info(f'[Learn phase 7/{n_phases}] - Ensemble analysis')
self.analyze_ensemble(enc_train_test)

# ------------------------ #
# Enable model partial fit AFTER it is trained and evaluated for performance with the appropriate train/dev/test splits.
# This assumes the predictor could continuously evolve, hence including reserved testing data may improve predictions.
# SET `json_ai.problem_definition.fit_on_all=False` TO TURN THIS BLOCK OFF.

# Update the mixers with partial fit
if self.problem_definition.fit_on_all and all([not m.trains_once for m in self.mixers]):
    log.info(f'[Learn phase 8/{n_phases}] - Adjustment on validation requested')
    self.adjust(enc_train_test["test"].data_frame, ConcatedEncodedDs([enc_train_test["train"],
                                                                      enc_train_test["dev"]]).data_frame,
                                                                      adjust_args={'learn_call': True})

self.feature_cache = dict()  # empty feature cache to avoid large predictor objects
"""
    learn_body = align(learn_body, 2)
    # ----------------- #
    # Predict Body
    # ----------------- #

    predict_body = f"""
self.mode = 'predict'
n_phases = 3 if self.pred_args.all_mixers else 4

if len(data) == 0:
    raise Exception("Empty input, aborting prediction. Please try again with some input data.")

self.pred_args = PredictionArguments.from_dict(args)

log.info(f'[Predict phase 1/{{n_phases}}] - Data preprocessing')
if self.problem_definition.ignore_features:
    log.info(f'Dropping features: {{self.problem_definition.ignore_features}}')
    data = data.drop(columns=self.problem_definition.ignore_features, errors='ignore')
for col in self.input_cols:
    if col not in data.columns:
        data[col] = [None] * len(data)

# Pre-process the data
data = self.preprocess(data)

# Featurize the data
log.info(f'[Predict phase 2/{{n_phases}}] - Feature generation')
encoded_ds = self.featurize({{"predict_data": data}})["predict_data"]
encoded_data = encoded_ds.get_encoded_data(include_target=False)

log.info(f'[Predict phase 3/{{n_phases}}] - Calling ensemble')

@timed
def _timed_call(encoded_ds):
    if self.pred_args.return_embedding:
        embedder = Embedder(self.target, mixers=list(), data=encoded_ds)
        df = embedder(encoded_ds, args=self.pred_args)
    else:
        df = self.ensemble(encoded_ds, args=self.pred_args)
    return df

df = _timed_call(encoded_ds)

if not(any(
            [self.pred_args.all_mixers,
             self.pred_args.return_embedding,
             self.problem_definition.embedding_only]
        )):
    log.info(f'[Predict phase 4/{{n_phases}}] - Analyzing output')
    df, global_insights = {call(json_ai.explainer)}
    self.global_insights = {{**self.global_insights, **global_insights}}

self.feature_cache = dict()  # empty feature cache to avoid large predictor objects

return df
"""

    predict_body = align(predict_body, 2)

    # ----------------- #
    # Test Body
    # ----------------- #
    test_body = """
preds = self.predict(data, args)
preds = preds.rename(columns={'prediction': self.target})
filtered = []

# filter metrics if not supported
for metric in metrics:
    # metric should be one of: an actual function, registered in the model class, or supported by the evaluator
    if not (callable(metric) or metric in self.accuracy_functions or metric in mdb_eval_accuracy_metrics):
        if strict:
            raise Exception(f'Invalid metric: {metric}')
        else:
            log.warning(f'Invalid metric: {metric}. Skipping...')
    else:
        filtered.append(metric)

metrics = filtered
try:
    labels = self.model_analysis.histograms[self.target]['x']
except:
    if strict:
        raise Exception('Label histogram not found')
    else:
        label_map = None  # some accuracy functions will crash without this, be mindful
scores = evaluate_accuracies(
                data,
                preds[self.target],
                self.target,
                metrics,
                ts_analysis=self.ts_analysis,
                labels=labels
            )

# TODO: remove once mdb_eval returns an actual list
scores = {k: [v] for k, v in scores.items() if not isinstance(v, list)}

return pd.DataFrame.from_records(scores)  # TODO: add logic to disaggregate per-mixer
"""

    test_body = align(test_body, 2)

    predictor_code = f"""
{IMPORTS}
{IMPORT_EXTERNAL_DIRS}

class Predictor(PredictorInterface):
    target: str
    mixers: List[BaseMixer]
    encoders: Dict[str, BaseEncoder]
    ensemble: BaseEnsemble
    mode: str

    def __init__(self):
        seed({json_ai.problem_definition.seed_nr})
        self.target = '{json_ai.problem_definition.target}'
        self.mode = 'inactive'
        self.problem_definition = ProblemDefinition.from_dict({json_ai.problem_definition.to_dict()})
        self.accuracy_functions = {json_ai.accuracy_functions}
        self.identifiers = {json_ai.identifiers}
        self.dtype_dict = {inline_dict(dtype_dict)}
        self.lightwood_version = '{lightwood_version}'
        self.pred_args = PredictionArguments()

        # Any feature-column dependencies
        self.dependencies = {inline_dict(json_ai.dependency_dict)}

        self.input_cols = [{input_cols}]

        # Initial stats analysis
        self.statistical_analysis = None
        self.ts_analysis = None
        self.runtime_log = dict()
        self.global_insights = dict()

        # Feature cache
        self.feature_cache = dict()

    @timed_predictor
    def analyze_data(self, data: pd.DataFrame) -> None:
        # Perform a statistical analysis on the unprocessed data
{analyze_data_body}

    @timed_predictor
    def preprocess(self, data: pd.DataFrame) -> pd.DataFrame:
        # Preprocess and clean data
{clean_body}

    @timed_predictor
    def split(self, data: pd.DataFrame) -> Dict[str, pd.DataFrame]:
        # Split the data into training/testing splits
{split_body}

    @timed_predictor
    def prepare(self, data: Dict[str, pd.DataFrame]) -> None:
        # Prepare encoders to featurize data
{prepare_body}

    @timed_predictor
    def featurize(self, split_data: Dict[str, pd.DataFrame]):
        # Featurize data into numerical representations for models
{feature_body}

    @timed_predictor
    def fit(self, enc_data: Dict[str, pd.DataFrame]) -> None:
        # Fit predictors to estimate target
{fit_body}

    @timed_predictor
    def fit_mixer(self, mixer, encoded_train_data, encoded_dev_data) -> None:
        mixer.fit(encoded_train_data, encoded_dev_data)

    @timed_predictor
    def analyze_ensemble(self, enc_data: Dict[str, pd.DataFrame]) -> None:
        # Evaluate quality of fit for the ensemble of mixers
{analyze_ensemble}

    @timed_predictor
    def learn(self, data: pd.DataFrame) -> None:
        if self.problem_definition.ignore_features:
            log.info(f'Dropping features: {{self.problem_definition.ignore_features}}')
            data = data.drop(columns=self.problem_definition.ignore_features, errors='ignore')
{learn_body}

    @timed_predictor
    def adjust(self, train_data: Union[EncodedDs, ConcatedEncodedDs, pd.DataFrame],
        dev_data: Optional[Union[EncodedDs, ConcatedEncodedDs, pd.DataFrame]] = None,
        adjust_args: Optional[dict] = None) -> None:
        # Update mixers with new information
{adjust_body}

    @timed_predictor
    def predict(self, data: pd.DataFrame, args: Dict = {{}}) -> pd.DataFrame:
{predict_body}

    def test(
        self, data: pd.DataFrame, metrics: list, args: Dict[str, object] = {{}}, strict: bool = False
        ) -> pd.DataFrame:
{test_body}
"""

    try:
        import black
    except Exception:
        black = None

    if black is not None:
        try:
            formatted_predictor_code = black.format_str(predictor_code, mode=black.FileMode())

            if type(_predictor_from_code(formatted_predictor_code)).__name__ == 'Predictor':
                predictor_code = formatted_predictor_code
            else:
                log.info('Black formatter output is invalid, predictor code might be a bit ugly')

        except Exception:
            log.info('Black formatter failed to run, predictor code might be a bit ugly')
    else:
        log.info('Unable to import black formatter, predictor code might be a bit ugly.')

    return predictor_code


def _module_from_code(code: str, module_name: str) -> ModuleType:
    """
    Create a python module (containing the generated ``Predictor`` class) from the code. This is both a python object and an associated temporary file on your filesystem

    :param code: The ``Predictor``'s code in text form
    :param module_name: The name of the newly created module

    :returns: A python module object
    """ # noqa
    dirname = tempfile.gettempdir()
    filename = os.urandom(24).hex() + str(time.time()).replace('.', '') + '.py'
    path = os.path.join(dirname, filename)
    if 'LIGHTWOOD_DEV_SAVE_TO' in os.environ:
        path = os.environ['LIGHTWOOD_DEV_SAVE_TO']

    with open(path, 'wb') as fp:
        fp.write(code.encode('utf-8'))
        spec = importlib.util.spec_from_file_location(module_name, fp.name)
        temp_module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = temp_module
        spec.loader.exec_module(temp_module)

    return temp_module


def _predictor_from_code(code: str):
    """
    :param code: The ``Predictor``'s code in text form

    :returns: A lightwood ``Predictor`` object
    """
    module_name = ''.join(random.choices(string.ascii_uppercase + string.digits, k=12))
    module_name += str(time.time()).replace('.', '')
    predictor = _module_from_code(code, module_name).Predictor()
    return predictor



import datetime
import calendar
import numpy as np
import pandas as pd  # @TODO: remove?
import torch
from lightwood.encoder.base import BaseEncoder
from collections.abc import Iterable
from lightwood.helpers.general import is_none


class DatetimeNormalizerEncoder(BaseEncoder):
    def __init__(self, is_target: bool = False, sinusoidal: bool = False):
        super().__init__(is_target)
        self.sinusoidal = sinusoidal
        self.fields = ['year', 'month', 'day', 'weekday', 'hour', 'minute', 'second']
        self.constants = {'year': 3000.0, 'month': 12.0, 'weekday': 7.0,
                          'hour': 24.0, 'minute': 60.0, 'second': 60.0}
        if self.sinusoidal:
            self.output_size = 2
        else:
            self.output_size = 7

    def prepare(self, priming_data):
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        self.is_prepared = True

    def encode(self, data):
        """
        :param data: # @TODO: receive a consistent data type here; currently either list of lists or pd.Series w/lists
        :return: encoded data
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if isinstance(data, pd.Series):
            data = data.values
        if not isinstance(data[0], Iterable):
            data = [data]

        ret = [self.encode_one(row) for row in data]

        return torch.Tensor(ret)

    def encode_one(self, data):
        """
        Encodes a list of unix_timestamps, or a list of tensors with unix_timestamps
        :param data: list of unix_timestamps (unix_timestamp resolution is seconds)
        :return: a list of vectors
        """
        ret = []
        for unix_timestamp in data:
            if is_none(unix_timestamp):
                if self.sinusoidal:
                    vector = [0, 1] * len(self.fields)
                else:
                    vector = [0] * len(self.fields)
            else:
                c = self.constants
                if isinstance(unix_timestamp, torch.Tensor):
                    unix_timestamp = unix_timestamp.item()
                date = datetime.datetime.fromtimestamp(unix_timestamp)
                day_constant = calendar.monthrange(date.year, date.month)[1]
                vector = [date.year / c['year'], date.month / c['month'], date.day / day_constant,
                          date.weekday() / c['weekday'], date.hour / c['hour'],
                          date.minute / c['minute'], date.second / c['second']]
                if self.sinusoidal:
                    vector = np.array([(np.sin(n), np.cos(n)) for n in vector]).flatten()

            ret.append(vector)

        return ret

    def decode(self, encoded_data, return_as_datetime=False):
        ret = []
        if len(encoded_data.shape) > 2 and encoded_data.shape[0] == 1:
            encoded_data = encoded_data.squeeze(0)

        for vector in encoded_data.tolist():
            ret.append(self.decode_one(vector, return_as_datetime=return_as_datetime))

        return ret

    def decode_one(self, vector, return_as_datetime=False):
        if sum(vector) == 0:
            decoded = None

        else:
            if self.sinusoidal:
                vector = list(map(lambda x: np.arcsin(x), vector))[::2]
            c = self.constants

            year = max(0, round(vector[0] * c['year']))
            month = max(1, min(12, round(vector[1] * c['month'])))
            day_constant = calendar.monthrange(year, month)[-1]
            day = max(1, min(round(vector[2] * day_constant), day_constant))
            hour = max(0, min(23, round(vector[4] * c['hour'])))
            minute = max(0, min(59, round(vector[5] * c['minute'])))
            second = max(0, min(59, round(vector[6] * c['second'])))

            dt = datetime.datetime(year=year, month=month, day=day, hour=hour,
                                   minute=minute, second=second)

            if return_as_datetime is True:
                decoded = dt
            else:
                decoded = round(dt.timestamp())

        return decoded



import copy
import time
import torch

import numpy as np
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.helpers.log import log


class Gym:

    def __init__(self, model, optimizer, scheduler, loss_criterion, device,
                 name=None, input_encoder=None, output_encoder=None):
        """
        Create an environment for training a pytorch machine learning model
        """
        self.device = device
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.loss_criterion = loss_criterion
        self.name = name
        self.input_encoder = input_encoder
        self.output_encoder = output_encoder

        self.best_model = None

    def fit(self, train_data_loader, test_data_loader, desired_error, max_time, callback=None,
            eval_every_x_epochs=1, max_unimproving_models=10, custom_train_func=None, custom_test_func=None):
        started = time.time()
        epoch = 0
        lowest_test_error = None
        last_test_error = None
        test_error_delta_buff = []

        keep_training = True

        while keep_training:
            epoch += 1
            running_loss = 0.0
            error = 0
            self.model = self.model.train()
            for i, data in enumerate(train_data_loader, 0):
                if custom_train_func is None:
                    input, real = data

                    with LightwoodAutocast():
                        if self.input_encoder is not None:
                            input = self.input_encoder(input)
                            if len(input.shape) < 2:
                                input = input.unsqueeze(-1)
                        if self.output_encoder is not None:
                            real = self.output_encoder(real)

                        input = input.to(self.device)
                        real = real.to(self.device)

                        predicted = self.model(input)
                        loss = self.loss_criterion(predicted, real)
                    loss.backward()
                    self.optimizer.step()

                    if self.scheduler is not None:
                        self.scheduler.step()

                    self.optimizer.zero_grad()
                else:
                    loss = custom_train_func(self.model, data, self)

                running_loss += loss.item()
                error = running_loss / (i + 1)

            # end of epoch checks
            log.debug(f'Categorical AutoEncoder train loss at epoch {epoch}: {round(error, 9)}')
            if epoch % eval_every_x_epochs == 0:
                if test_data_loader is not None:
                    test_running_loss = 0.0
                    test_error = 0
                    self.model = self.model.eval()
                    real_buff = []
                    predicted_buff = []

                    for i, data in enumerate(test_data_loader, 0):
                        if custom_test_func is None:
                            input, real = data

                            if self.input_encoder is not None:
                                input = self.input_encoder(input)
                            if self.output_encoder is not None:
                                real = self.output_encoder(real)

                            input = input.to(self.device)
                            real = real.to(self.device)

                            with torch.no_grad():
                                predicted = self.model(input)

                            real_buff.append(real.tolist())
                            predicted_buff.append(predicted.tolist())

                            loss = self.loss_criterion(predicted, real)
                        else:
                            with torch.no_grad():
                                loss = custom_test_func(self.model, data, self)

                        log.debug(f'Categorical AutoEncoder val loss at epoch {epoch}: {round(loss, 9)}')
                        test_running_loss += loss.item()
                        test_error = test_running_loss / (i + 1)
                else:
                    test_error = error
                    real_buff = None
                    predicted_buff = None

                if lowest_test_error is None or test_error < lowest_test_error:
                    lowest_test_error = test_error
                    self.best_model = copy.deepcopy(self.model).to('cpu')

                if last_test_error is None:
                    test_error_delta_buff.append(0)
                else:
                    test_error_delta_buff.append(last_test_error - test_error)

                last_test_error = test_error

                if (time.time() - started) > max_time:
                    keep_training = False

                if lowest_test_error < desired_error:
                    keep_training = False

                if len(test_error_delta_buff) >= max_unimproving_models:
                    delta_mean = np.mean(test_error_delta_buff[-max_unimproving_models:])
                    if delta_mean <= 0:
                        keep_training = False
                if callback is not None:
                    callback(test_error, real_buff, predicted_buff)

        return self.best_model, lowest_test_error, int(time.time() - started)



from typing import List
import pandas as pd

from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs


class IdentityEnsemble(BaseEnsemble):
    """
    This ensemble performs no aggregation. User can define an "active mixer" and calling the ensemble will call said mixer. 
    
    Ideal for use cases with single mixers where (potentially expensive) evaluation runs are done internally, as in `BestOf`.
    """  # noqa

    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, args: PredictionArguments) -> None:
        super().__init__(target, mixers, data=data)
        self._active_mixer = 0
        single_row_ds = EncodedDs(data.encoders, data.data_frame.iloc[[0]], data.target)
        _ = self.mixers[self._active_mixer](single_row_ds, args)['prediction']  # prime mixer for storage, needed because NHitsMixer.model (neuralforecast.NHITS) is not serializable without this, oddly enough. Eventually, check this again and remove if possible!  # noqa
        self.prepared = True

    def __call__(self, ds: EncodedDs, args: PredictionArguments = None) -> pd.DataFrame:
        assert self.prepared
        mixer = self.mixers[self.active_mixer]
        return mixer(ds, args=args)

    @property
    def active_mixer(self):
        return self._active_mixer

    @active_mixer.setter
    def active_mixer(self, idx):
        assert 0 <= idx < len(self.mixers), f'The ensemble has {len(self.mixers)} mixers, please provide a valid index.'
        self._active_mixer = idx



"""
Reserved constants for Lightwood.
"""

_UNCOMMON_WORD = '__mdb_unknown_cat'
_UNCOMMON_TOKEN = 0

# For custom modules, we create a module loader with necessary imports below
IMPORT_EXTERNAL_DIRS = """
for import_dir in [os.path.join(os.path.expanduser('~/lightwood_modules'), lightwood_version.replace('.', '_')), os.path.join('/etc/lightwood_modules', lightwood_version.replace('.', '_'))]:
    if os.path.exists(import_dir) and os.access(import_dir, os.R_OK):
        for file_name in list(os.walk(import_dir))[0][2]:
            if file_name[-3:] != '.py':
                continue
            mod_name = file_name[:-3]
            loader = importlib.machinery.SourceFileLoader(mod_name,
                                                          os.path.join(import_dir, file_name))
            module = ModuleType(loader.name)
            loader.exec_module(module)
            sys.modules[mod_name] = module
            exec(f'import {mod_name}')
""" # noqa

IMPORTS = """
import lightwood
from lightwood import __version__ as lightwood_version
from lightwood.analysis import *
from lightwood.api import *
from lightwood.data import *
from lightwood.encoder import *
from lightwood.ensemble import *
from lightwood.helpers.device import *
from lightwood.helpers.general import *
from lightwood.helpers.ts import *
from lightwood.helpers.log import *
from lightwood.helpers.numeric import *
from lightwood.helpers.parallelism import *
from lightwood.helpers.seed import *
from lightwood.helpers.text import *
from lightwood.helpers.torch import *
from lightwood.mixer import *

from dataprep_ml.insights import statistical_analysis
from dataprep_ml.cleaners import cleaner
from dataprep_ml.splitters import splitter
from dataprep_ml.imputers import *

from mindsdb_evaluator import evaluate_accuracies
from mindsdb_evaluator.accuracy import __all__ as mdb_eval_accuracy_metrics

import pandas as pd
from typing import Dict, List, Union, Optional
import os
from types import ModuleType
import importlib.machinery
import sys
import time
"""



from typing import Dict, List

from lightwood.encoder.base import BaseEncoder
from lightwood.mixer.neural import Neural
from lightwood.mixer.helpers.qclassic_net import QClassicNet
from lightwood.api.types import TimeseriesSettings


class QClassic(Neural):
    # wrapper class to be combined with Neural class when performance stabilizes
    def __init__(
            self, stop_after: float, target: str, dtype_dict: Dict[str, str],
            input_cols: List[str],
            timeseries_settings: TimeseriesSettings, target_encoder: BaseEncoder, net: str, fit_on_dev: bool,
            search_hyperparameters: bool):
        super().__init__(stop_after, target, dtype_dict,
                         input_cols, timeseries_settings, target_encoder,
                         net, fit_on_dev, search_hyperparameters)

        quantum_nets = {"QClassic": QClassicNet}
        self.net_class = quantum_nets.get(net, QClassicNet)



import math
from typing import Union, Dict
from copy import deepcopy as dc

import torch
import numpy as np
import pandas as pd
from type_infer.dtype import dtype

from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.general import is_none


class NumericEncoder(BaseEncoder):
    """
    The numeric encoder takes numbers (float or integer) and converts it into tensors of the form:
    ``[0 if the number is none, otherwise 1, 1 if the number is positive, otherwise 0, natural_log(abs(number)), number/absolute_mean]``

    This representation is: ``[1 if the number is positive, otherwise 0, natural_log(abs(number)), number/absolute_mean]]`` if encoding target values, since target values can't be none.

    The ``absolute_mean`` is computed in the ``prepare`` method and is just the mean of the absolute values of all numbers feed to prepare (which are not none)

    ``none`` stands for any number that is an actual python ``None`` value or any sort of non-numeric value (a string, nan, inf)
    """  # noqa

    def __init__(self, data_type: dtype = None,
                 target_weights: Dict[float, float] = None,
                 is_target: bool = False,
                 positive_domain: bool = False):
        """
        :param data_type: The data type of the number (integer, float, quantity)
        :param target_weights: a dictionary of weights to use on the examples.
        :param is_target: Indicates whether the encoder refers to a target column or feature column (True==target)
        :param positive_domain: Forces the encoder to always output positive values
        """
        super().__init__(is_target)
        self._abs_mean = None
        self.positive_domain = positive_domain
        self.decode_log = False
        self.output_size = 4 if not self.is_target else 3

        # Weight-balance info if encoder represents target
        self.target_weights = None
        self.index_weights = None
        if self.is_target and target_weights is not None:
            self.target_weights = dc(target_weights)
            self.index_weights = torch.tensor(list(self.target_weights.values()))

    def prepare(self, priming_data: pd.Series):
        """
        "NumericalEncoder" uses a rule-based form to prepare results on training (priming) data. The averages etc. are taken from this distribution.

        :param priming_data: an iterable data structure containing numbers numbers which will be used to compute the values used for normalizing the encoded representations
        """  # noqa
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        self._abs_mean = priming_data.abs().mean()
        self.is_prepared = True

    def encode(self, data: Union[np.ndarray, pd.Series]):
        """
        :param data: A pandas series or numpy array containing the numbers to be encoded
        :returns: A torch tensor with the representations of each number
        """
        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if isinstance(data, pd.Series):
            data = data.values

        inp_data = np.nan_to_num(data.astype(float), nan=0, posinf=np.finfo(np.float32).max,
                                 neginf=np.finfo(np.float32).min)  # noqa
        if not self.positive_domain:
            sign = np.vectorize(self._sign_fn, otypes=[float])(inp_data)
        else:
            sign = np.zeros(len(data))
        log_value = np.vectorize(self._log_fn, otypes=[float])(inp_data)
        log_value = np.nan_to_num(log_value, nan=0, posinf=20, neginf=-20)

        norm = np.vectorize(self._norm_fn, otypes=[float])(inp_data)
        norm = np.nan_to_num(norm, nan=0, posinf=20, neginf=-20)

        if self.is_target:
            components = [sign, log_value, norm]
        else:
            nones = np.vectorize(self._none_fn, otypes=[float])(data)
            components = [sign, log_value, norm, nones]

        return torch.Tensor(np.asarray(components)).T

    @staticmethod
    def _sign_fn(x: float) -> float:
        return 0 if x < 0 else 1

    @staticmethod
    def _log_fn(x: float) -> float:
        return math.log(abs(x)) if abs(x) > 0 else -20

    def _norm_fn(self, x: float) -> float:
        return x / self._abs_mean

    @staticmethod
    def _none_fn(x: float) -> float:
        return 1 if is_none(x) else 0

    def decode(self, encoded_values: torch.Tensor, decode_log: bool = None) -> list:
        """
        :param encoded_values: The encoded values to decode into single numbers
        :param decode_log: Whether to decode the ``log`` or ``linear`` part of the representation, since the encoded vector contains both a log and a linear part

        :returns: The decoded array
        """  # noqa

        if not self.is_prepared:
            raise Exception('You need to call "prepare" before calling "encode" or "decode".')

        if decode_log is None:
            decode_log = self.decode_log

        # force = True prevents side effects on the original encoded_values
        ev = encoded_values.numpy(force=True)

        # set "divergent" value as default (note: finfo.max() instead of pow(10, 63))
        ret = np.full((ev.shape[0],), dtype=float, fill_value=np.finfo(np.float64).max)

        # `none` filter (if not a target column)
        if not self.is_target:
            mask_none = ev[:, -1] == 1
            ret[mask_none] = np.nan
        else:
            mask_none = np.zeros_like(ret)

        # sign component
        sign = np.ones(ev.shape[0], dtype=float)
        mask_sign = ev[:, 0] < 0.5
        sign[mask_sign] = -1

        # real component
        if decode_log:
            real_value = np.exp(ev[:, 1]) * sign
            overflow_mask = ev[:, 1] >= 63
            real_value[overflow_mask] = 10 ** 63
            valid_mask = ~overflow_mask
        else:
            real_value = ev[:, 2] * self._abs_mean
            valid_mask = np.ones_like(real_value, dtype=bool)

        # final filters
        if self.positive_domain:
            real_value = abs(real_value)

        ret[valid_mask] = real_value[valid_mask]

        # set nan back to None
        if mask_none.sum() > 0:
            ret = ret.astype(object)
            ret[mask_none] = None

        return ret.tolist()  # TODO: update signature on BaseEncoder and replace all encs to return ndarrays

    def get_weights(self, label_data):
        # get a sorted list of intervals to assign weights. Keys are the interval edges.
        target_weight_keys = np.array(list(self.target_weights.keys()))
        target_weight_values = np.array(list(self.target_weights.values()))
        sorted_indices = np.argsort(target_weight_keys)

        # get sorted arrays for vector numpy operations
        target_weight_keys = target_weight_keys[sorted_indices]
        target_weight_values = target_weight_values[sorted_indices]

        # find the indices of the bins according to the keys. clip to the length of the weight values (search sorted
        # returns indices from 0 to N with N = len(target_weight_keys).
        assigned_target_weight_indices = np.clip(a=np.searchsorted(target_weight_keys, label_data),
                                                 a_min=0,
                                                 a_max=len(target_weight_keys) - 1).astype(np.int32)

        return target_weight_values[assigned_target_weight_indices]




import math
import torch
from torch.optim.optimizer import Optimizer


class Ranger(Optimizer):
    def __init__(
            self, params, lr=0.0005, alpha=0.5, k=5, N_sma_threshold=5, betas=(0.9, 0.999),
            eps=1e-5, weight_decay=0.000):
        # parameter checks
        if not 0.0 <= alpha <= 1.0:
            raise ValueError(f'Invalid slow update rate: {alpha}')
        if not 1 <= k:
            raise ValueError(f'Invalid lookahead steps: {k}')
        if not lr > 0:
            raise ValueError(f'Invalid Learning Rate: {lr}')
        if not eps > 0:
            raise ValueError(f'Invalid eps: {eps}')

        # parameter comments:
        # beta1 (momentum) of .95 seems to work better than .90...
        # N_sma_threshold of 5 seems better in testing than 4.
        # In both cases, worth testing on your dataset (.90 vs .95, 4 vs 5) to make sure which works best for you.
        # @TODO Implement the above testing with AX ^

        # prep defaults and init torch.optim base
        defaults = dict(lr=lr, alpha=alpha, k=k, betas=betas,
                        N_sma_threshold=N_sma_threshold, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)

        # Since we keep LR the same for all param groups,
        # store it here for now for quick&easy access if we want to know it
        self.lr = lr

        # adjustable threshold
        self.N_sma_threshold = N_sma_threshold

        # look ahead params
        self.initial_lr = lr
        self.alpha = alpha
        self.k = k

        # radam buffer for state
        self.radam_buffer = [[None, None, None] for ind in range(10)]

    def __setstate__(self, state):
        super(Ranger, self).__setstate__(state)

    def step(self, closure=None):
        loss = None

        if closure is not None:
            loss = closure()

        # Evaluate averages and grad, update param tensors
        for group in self.param_groups:
            for p in group['params']:
                if p.grad is not None:
                    grad = p.grad.data.float()
                    if grad.is_sparse:
                        raise RuntimeError('Ranger optimizer does not support sparse gradients')

                    p_data_fp32 = p.data.float()

                    state = self.state[p]  # get state dict for this param

                    # On the first run initialize the dictionary for each weight group
                    if len(state) == 0:
                        state['step'] = 0
                        state['exp_avg'] = torch.zeros_like(p_data_fp32)
                        state['exp_avg_sq'] = torch.zeros_like(p_data_fp32)

                        # look ahead weight storage now in state dict
                        state['slow_buffer'] = torch.empty_like(p.data)
                        state['slow_buffer'].copy_(p.data)
                    # @TODO Couldn't this branch happen after the if above is entered
                    # in thus replacing torch.zero_like) ??
                    else:
                        state['exp_avg'] = state['exp_avg'].type_as(p_data_fp32)
                        state['exp_avg_sq'] = state['exp_avg_sq'].type_as(p_data_fp32)

                # begin computations
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                # compute variance mov avg
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                # compute mean moving avg
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                state['step'] += 1

                buffered = self.radam_buffer[int(state['step'] % 10)]

                if state['step'] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state['step']
                    beta2_t = beta2 ** state['step']
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state['step'] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma
                    if N_sma > self.N_sma_threshold:
                        step_size = math.sqrt((1 - beta2_t) * (N_sma - 4) / (N_sma_max - 4) * (N_sma - 2) /
                                              N_sma * N_sma_max / (N_sma_max - 2)) / (1 - beta1 ** state['step'])
                    else:
                        step_size = 1.0 / (1 - beta1 ** state['step'])
                    buffered[2] = step_size

                if group['weight_decay'] != 0:
                    p_data_fp32.add_(-group['weight_decay'] * group['lr'], p_data_fp32)

                if N_sma > self.N_sma_threshold:
                    denom = exp_avg_sq.sqrt().add_(group['eps'])
                    p_data_fp32.addcdiv_(exp_avg, denom, value=-step_size * group['lr'])
                else:
                    p_data_fp32.add_(exp_avg, alpha=-step_size * group['lr'])

                p.data.copy_(p_data_fp32)

                # integrated look ahead...
                # we do it at the param level instead of group level
                if state['step'] % group['k'] == 0:
                    slow_p = state['slow_buffer']
                    # Find the interpolated weight between the slower buffer (the weight `k` steps ago)
                    # and the current weight, set that as the state for RAdam
                    slow_p.add_(p.data - slow_p, alpha=self.alpha)
                    p.data.copy_(slow_p)

        return loss



import inspect
from copy import deepcopy
from typing import Dict, Tuple, Optional
from types import SimpleNamespace

import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from type_infer.dtype import dtype
from lightwood.api.types import PredictionArguments
from lightwood.helpers.ts import add_tn_num_conf_bounds, add_tn_cat_conf_bounds, get_ts_groups

from lightwood.data import EncodedDs
from lightwood.analysis.base import BaseAnalysisBlock
from lightwood.analysis.nc.norm import Normalizer
from lightwood.analysis.nc.icp import IcpRegressor, IcpClassifier, IcpTSRegressor
from lightwood.analysis.nc.base import CachedRegressorAdapter, CachedClassifierAdapter, CachedTSAdapter
from lightwood.analysis.nc.nc import BoostedAbsErrorErrFunc, RegressorNc, ClassifierNc, MarginErrFunc, TSNc, \
    TSAbsErrorErrFunc
from lightwood.analysis.nc.util import clean_df, set_conf_range, get_numeric_conf_range, \
    get_categorical_conf, get_anomalies, get_ts_conf_range


class ICP(BaseAnalysisBlock):
    """ Confidence estimation block, uses inductive conformal predictors (ICPs) for model agnosticity """

    def __init__(self,
                 confidence_normalizer: Optional[bool] = False,
                 fixed_significance: Optional[float] = None,
                 deps: Optional[tuple] = tuple()
                 ):
        super().__init__(deps=deps)
        self.fixed_significance = fixed_significance
        self.confidence_normalizer = confidence_normalizer
        self.validation_size = 100  # determines size of nonconformity score arrays (has sizable impact in runtime)

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        ns = SimpleNamespace(**kwargs)

        data_type = ns.dtype_dict[ns.target]
        output = {'icp': {'__mdb_active': False}}

        if 'confidence' in ns.normal_predictions.columns:
            #  bypass calibrator if model already outputs confidence
            output['result_df'] = ns.normal_predictions[['confidence', 'lower', 'upper']]
            return {**info, **output}

        fit_params = {'horizon': ns.tss.horizon or 0, 'columns_to_ignore': []}
        fit_params['columns_to_ignore'].extend([f'timestep_{i}' for i in range(1, fit_params['horizon'])])

        if ns.is_classification:
            if ns.predictor.supports_proba:
                all_cat_cols = [col for col in ns.normal_predictions.columns
                                if '__mdb_proba' in col and '__mdb_unknown_cat' not in col]
                all_classes = np.array([col.replace('__mdb_proba_', '') for col in all_cat_cols])
            else:
                class_keys = sorted(ns.encoded_val_data.encoders[ns.target].rev_map.keys())
                all_classes = np.array([ns.encoded_val_data.encoders[ns.target].rev_map[idx] for idx in class_keys])

            if data_type != dtype.tags:
                enc = OneHotEncoder(sparse=False, handle_unknown='ignore')
                enc.fit(all_classes.reshape(-1, 1))
                output['label_encoders'] = enc  # needed to repr cat labels inside nonconformist
            else:
                output['label_encoders'] = None

            adapter = CachedClassifierAdapter
            nc_function = MarginErrFunc()
            nc_class = ClassifierNc
            icp_class = IcpClassifier
        elif ns.is_multi_ts:
            adapter = CachedTSAdapter
            nc_function = TSAbsErrorErrFunc(horizon_length=ns.tss.horizon)
            nc_class = TSNc
            icp_class = IcpTSRegressor
        else:
            adapter = CachedRegressorAdapter
            nc_function = BoostedAbsErrorErrFunc()
            nc_class = RegressorNc
            icp_class = IcpRegressor

        result_df = pd.DataFrame()

        if ns.is_numerical or (ns.is_classification and data_type != dtype.tags):
            model = adapter(ns.predictor)

            norm_params = {'target': ns.target, 'dtype_dict': ns.dtype_dict, 'predictor': ns.predictor,
                           'encoders': ns.encoded_val_data.encoders, 'is_multi_ts': ns.is_multi_ts, 'stop_after': 1e2}
            if self.confidence_normalizer:
                normalizer = Normalizer(fit_params=norm_params)
                normalizer.fit(ns.train_data)
                normalizer.prediction_cache = normalizer(ns.encoded_val_data, args=PredictionArguments())
            else:
                normalizer = None

            # instance the ICP
            nc = nc_class(model, nc_function, normalizer=normalizer)
            if 'horizon_length' in inspect.signature(icp_class).parameters:
                icp = icp_class(nc, horizon_length=ns.tss.horizon, cal_size=self.validation_size)
            else:
                icp = icp_class(nc, cal_size=self.validation_size)

            output['icp']['__default'] = icp
            icp_df = deepcopy(ns.data)

            # setup prediction cache to avoid additional .predict() calls
            try:
                pred_is_list = isinstance(ns.normal_predictions['prediction'][0], list)
            except KeyError:
                pred_is_list = False

            if ns.is_classification:
                if ns.predictor.supports_proba:
                    icp.nc_function.model.prediction_cache = ns.normal_predictions[all_cat_cols].values
                else:
                    if ns.is_multi_ts:
                        preds = np.array([p[0] for p in ns.normal_predictions['prediction']])
                    else:
                        preds = ns.normal_predictions['prediction'].values
                    predicted_classes = output['label_encoders'].transform(preds.reshape(-1, 1))  # inflate OHE
                    icp.nc_function.model.prediction_cache = predicted_classes

            elif ns.is_multi_ts or pred_is_list:
                icp.nc_function.model.prediction_cache = np.array(
                    [np.array(p) for p in ns.normal_predictions['prediction']])
            else:
                icp.nc_function.model.prediction_cache = np.array(ns.normal_predictions['prediction'])

            if not ns.is_classification:
                output['df_target_stddev'] = {'__default': ns.stats_info.df_target_stddev}

            # fit additional ICPs in time series tasks with grouped columns
            if ns.tss.is_timeseries and ns.tss.group_by:
                # generate a multiindex
                midx = pd.MultiIndex.from_frame(icp_df[[*ns.tss.group_by, f'__mdb_original_{ns.tss.order_by}']])
                icp_df.index = midx
                result_df.index = midx

                # create an ICP for each possible group
                group_info = ns.data[ns.tss.group_by].to_dict('list')
                all_group_combinations = get_ts_groups(ns.data, ns.tss)
                all_group_combinations.remove('__default')
                output['icp']['__mdb_groups'] = all_group_combinations
                output['icp']['__mdb_group_keys'] = [x for x in group_info.keys()]

                for combination in all_group_combinations:
                    output['icp'][tuple(combination)] = deepcopy(icp)

            # calibrate ICP
            icp_df = deepcopy(ns.data)
            icp_df, y = clean_df(icp_df, ns, output.get('label_encoders', None))
            if ns.tss.is_timeseries and ns.tss.group_by:
                icp_df.index = midx
            output['icp']['__default'].index = icp_df.columns
            output['icp']['__default'].calibrate(icp_df.values, y)

            # get confidence estimation for validation dataset
            conf, ranges = set_conf_range(
                icp_df, icp, ns.dtype_dict[ns.target],
                output, positive_domain=ns.stats_info.positive_domain, significance=self.fixed_significance)
            if not ns.is_classification:
                result_df = pd.DataFrame(index=icp_df.index, columns=['confidence', 'lower', 'upper'], dtype=float)
                result_df.loc[icp_df.index, 'lower'] = ranges[:, 0]
                result_df.loc[icp_df.index, 'upper'] = ranges[:, 1]
            else:
                result_df = pd.DataFrame(index=icp_df.index, columns=['confidence'], dtype=float)

            result_df.loc[icp_df.index, 'confidence'] = conf

            # calibrate additional grouped ICPs
            if ns.tss.is_timeseries and ns.tss.group_by:
                icps = output['icp']
                group_keys = icps['__mdb_group_keys']

                # add all predictions to DF
                icps_df = deepcopy(ns.data)
                midx = pd.MultiIndex.from_frame(icps_df[[*ns.tss.group_by, f'__mdb_original_{ns.tss.order_by}']])
                icps_df.index = midx
                if ns.is_multi_ts or pred_is_list:
                    icps_df[f'__predicted_{ns.target}'] = [np.array(p) for p in ns.normal_predictions['prediction']]
                else:
                    icps_df[f'__predicted_{ns.target}'] = np.array(ns.normal_predictions['prediction'])

                for group in icps['__mdb_groups']:
                    icp_df = icps_df

                    # filter irrelevant rows for each group combination
                    icp_df['__mdb_norm_index'] = np.arange(len(icp_df))
                    for key, val in zip(group_keys, group):
                        icp_df = icp_df[icp_df[key] == val]

                    if icps[tuple(group)].nc_function.normalizer is not None:
                        group_normalizer = icps[tuple(group)].nc_function.normalizer
                        norm_input_df = ns.encoded_val_data.data_frame.iloc[icp_df.pop('__mdb_norm_index')]
                        norm_input = EncodedDs(ns.encoded_val_data.encoders, norm_input_df, ns.target)
                        norm_cache = group_normalizer(norm_input, args=PredictionArguments())
                        icp_df[f'__norm_{ns.target}'] = norm_cache

                    # save relevant predictions in the caches, then calibrate the ICP
                    pred_cache = icp_df.pop(f'__predicted_{ns.target}').values
                    if ns.is_multi_ts and ns.is_classification:
                        pred_cache = output['label_encoders'].transform([[p[0] for p in pred_cache]])
                    elif ns.is_multi_ts:
                        pred_cache = np.array([np.array(p) for p in pred_cache])
                    elif ns.is_classification:
                        pred_cache = output['label_encoders'].transform(pred_cache.reshape(-1, 1))

                    icps[tuple(group)].nc_function.model.prediction_cache = pred_cache
                    icp_df, y = clean_df(icp_df, ns, output.get('label_encoders', None))
                    if icps[tuple(group)].nc_function.normalizer is not None:
                        icps[tuple(group)].nc_function.normalizer.prediction_cache = icp_df.pop(
                            f'__norm_{ns.target}').values

                    icps[tuple(group)].index = icp_df.columns  # important at inference time
                    icps[tuple(group)].calibrate(icp_df.values, y)

                    # save training std() for bounds width selection
                    if not ns.is_classification:
                        icp_train_df = ns.data
                        for key, val in zip(group_keys, group):
                            icp_train_df = icp_train_df[icp_train_df[key] == val]
                        y_train = icp_train_df[ns.target].values
                        output['df_target_stddev'][tuple(group)] = y_train.std()

                    # get bounds for relevant rows in validation dataset
                    conf, group_ranges = set_conf_range(
                        icp_df, icps[tuple(group)],
                        ns.dtype_dict[ns.target],
                        output, group=tuple(group),
                        positive_domain=ns.stats_info.positive_domain, significance=self.fixed_significance)
                    # save group bounds
                    if not ns.is_classification:
                        result_df.loc[icp_df.index, 'lower'] = group_ranges[:, 0]
                        result_df.loc[icp_df.index, 'upper'] = group_ranges[:, 1]

                    result_df.loc[icp_df.index, 'confidence'] = conf

            # consolidate all groups here
            output['icp']['__mdb_active'] = True

        result_df.index = ns.data.index
        output['result_df'] = result_df

        info = {**info, **output}
        return info

    def explain(self, row_insights: pd.DataFrame, global_insights: Dict[str, object],
                **kwargs) -> Tuple[pd.DataFrame, Dict[str, object]]:
        ns = SimpleNamespace(**kwargs)

        is_categorical = ns.target_dtype in (dtype.binary, dtype.categorical, dtype.cat_array, dtype.cat_tsarray)
        is_numerical = ns.target_dtype in (dtype.integer, dtype.float,
                                           dtype.quantity, dtype.num_array, dtype.num_tsarray)
        is_multi_ts = ns.tss.is_timeseries and ns.tss.horizon > 1
        is_anomaly_task = is_numerical and ns.tss.is_timeseries and ns.anomaly_detection

        if 'confidence' in ns.predictions.columns:
            # bypass calibrator if model already outputs confidence
            row_insights['prediction'] = ns.predictions['prediction']
            if 'upper' in ns.predictions.columns and 'lower' in ns.predictions.columns:
                row_insights['upper'] = ns.predictions['upper']
                row_insights['lower'] = ns.predictions['lower']

            if not isinstance(ns.predictions['confidence'].iloc[0], list) and ns.tss.horizon > 1:
                row_insights['confidence'] = ns.predictions['confidence'].astype(object)
                row_insights['confidence'] = row_insights['confidence'].apply(
                    lambda x: [x for _ in range(ns.tss.horizon)])
            else:
                row_insights['confidence'] = ns.predictions['confidence']
            return self._formatted(row_insights, global_insights, ns, is_numerical)

        if ns.analysis['icp']['__mdb_active']:
            icp_X = deepcopy(ns.data)

            # replace observed data w/predictions
            preds = ns.predictions['prediction']
            if is_multi_ts and is_numerical:
                preds = np.array([np.array(p) for p in preds])

                for col in [f'timestep_{i}' for i in range(1, ns.tss.horizon)]:
                    if col in icp_X.columns:
                        icp_X.pop(col)  # erase ignorable columns

                target_cols = [ns.target_name] + [f'{ns.target_name}_timestep_{i}' for i in range(1, ns.tss.horizon)]
                icp_X[target_cols] = preds
            elif is_multi_ts and is_categorical:
                preds = [p[0] for p in preds]
                icp_X[ns.target_name] = preds
            else:
                icp_X[ns.target_name] = preds

            if (is_numerical or is_categorical) and ns.analysis['icp'].get('__mdb_active', False):
                base_icp = ns.analysis['icp']['__default']
                # reorder DF index
                index = base_icp.index.values
                if ns.target_name not in index:
                    if is_multi_ts:
                        index = np.array(list(index) + [ns.target_name] +
                                         [f'{ns.target_name}_timestep_{i}' for i in range(1, ns.tss.horizon)])
                    else:
                        index = np.append(index, ns.target_name)
                icp_X = icp_X.reindex(columns=index)  # important, else bounds can be invalid

                # only one normalizer, even if it's a grouped time series task
                normalizer = base_icp.nc_function.normalizer
                if normalizer:
                    normalizer.prediction_cache = normalizer(ns.encoded_data, args=PredictionArguments)
                    icp_X['__mdb_selfaware_scores'] = normalizer.prediction_cache

                # get ICP predictions
                if is_multi_ts:
                    result_cols = ['significance', 'lower', 'upper', ] + \
                                  [f'lower_timestep_{i}' for i in range(1, ns.tss.horizon)] + \
                                  [f'upper_timestep_{i}' for i in range(1, ns.tss.horizon)] + \
                                  [f'significance_timestep_{i}' for i in range(1, ns.tss.horizon)]
                elif is_numerical:
                    result_cols = ['lower', 'upper', 'significance']
                else:
                    result_cols = ['significance']
                result = pd.DataFrame(index=icp_X.index, columns=result_cols)

                # base ICP
                X = deepcopy(icp_X)
                # Calling `values` multiple times increased runtime of this function; referenced var is faster
                icp_values = X.values

                # get all possible ranges
                if is_numerical:
                    base_icp.nc_function.model.prediction_cache = preds
                    all_confs = base_icp.predict(icp_values)

                # categorical
                else:
                    predicted_proba = True if any(['__mdb_proba' in col for col in ns.predictions.columns]) else False
                    if predicted_proba:
                        all_cat_cols = [col for col in ns.predictions.columns
                                        if '__mdb_proba' in col and '__mdb_unknown_cat' not in col]
                        class_dists = ns.predictions[all_cat_cols].values
                        for icol, cat_col in enumerate(all_cat_cols):
                            row_insights.loc[X.index, cat_col] = class_dists[:, icol]
                    else:
                        ohe_enc = ns.analysis['label_encoders']
                        class_dists = ohe_enc.transform(np.array([p[0] for p in preds]).reshape(-1, 1))

                    base_icp.nc_function.model.prediction_cache = class_dists

                    all_ranges = np.array([base_icp.predict(icp_values)])
                    all_confs = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)

                # convert (B, 2, 99) into (B, 2) given width or error rate constraints
                if is_multi_ts and is_numerical:
                    significances, confs = get_ts_conf_range(all_confs,
                                                             df_target_stddev=ns.analysis['df_target_stddev'],
                                                             positive_domain=ns.positive_domain,
                                                             fixed_conf=ns.pred_args.fixed_confidence)

                    result = self._ts_assign_confs(result, X, confs, significances, ns.tss)

                elif is_numerical:
                    significances, confs = get_numeric_conf_range(all_confs,
                                                                  df_target_stddev=ns.analysis['df_target_stddev'],
                                                                  positive_domain=ns.positive_domain,
                                                                  fixed_conf=ns.pred_args.fixed_confidence)
                    result.loc[X.index, 'lower'] = confs[:, 0]
                    result.loc[X.index, 'upper'] = confs[:, 1]
                    result.loc[X.index, 'significance'] = significances

                else:
                    significances = get_categorical_conf(all_confs)
                    result.loc[X.index, 'significance'] = significances.flatten()

                # grouped time series, we replace bounds in rows that have a trained ICP
                if ns.analysis['icp'].get('__mdb_groups', False):
                    icps = ns.analysis['icp']
                    group_keys = icps['__mdb_group_keys']

                    for group in icps['__mdb_groups']:
                        icp = icps[tuple(group)]

                        # check ICP has calibration scores
                        if icp.cal_scores[0].shape[0] > 0:

                            # filter rows by group
                            X = deepcopy(icp_X)
                            for key, val in zip(group_keys, group):
                                X = X[X[key] == val]

                            if X.size > 0:
                                # set ICP caches
                                if is_multi_ts and is_numerical:
                                    target_cols = [ns.target_name] + [f'{ns.target_name}_timestep_{i}'
                                                                      for i in range(1, ns.tss.horizon)]
                                    icp.nc_function.model.prediction_cache = X[target_cols].values
                                    [X.pop(col) for col in target_cols]
                                elif is_multi_ts and is_categorical:
                                    ohe_enc = ns.analysis['label_encoders']
                                    preds = X.pop(ns.target_name).values
                                    pred_cache = ohe_enc.transform(np.array([p[0] for p in preds]).reshape(-1, 1))
                                    icp.nc_function.model.prediction_cache = pred_cache
                                else:
                                    icp.nc_function.model.prediction_cache = X.pop(ns.target_name).values
                                if icp.nc_function.normalizer:
                                    icp.nc_function.normalizer.prediction_cache = X.pop('__mdb_selfaware_scores').values

                                # predict and get confidence level given width or error rate constraints
                                if is_multi_ts and is_numerical:
                                    all_confs = icp.predict(X.values)
                                    fixed_conf = ns.pred_args.fixed_confidence
                                    significances, confs = get_ts_conf_range(
                                        all_confs,
                                        df_target_stddev=ns.analysis['df_target_stddev'],
                                        positive_domain=ns.positive_domain,
                                        group=tuple(group),
                                        fixed_conf=fixed_conf
                                    )
                                    result = self._ts_assign_confs(result, X, confs, significances, ns.tss)

                                elif is_numerical:
                                    all_confs = icp.predict(X.values)
                                    fixed_conf = ns.pred_args.fixed_confidence
                                    significances, confs = get_numeric_conf_range(
                                        all_confs,
                                        df_target_stddev=ns.analysis['df_target_stddev'],
                                        positive_domain=ns.positive_domain,
                                        group=tuple(group),
                                        fixed_conf=fixed_conf
                                    )

                                    # only replace where grouped ICP is more informative (i.e. tighter)
                                    if fixed_conf is None:
                                        default_widths = result.loc[X.index, 'upper'] - result.loc[X.index, 'lower']
                                        grouped_widths = np.subtract(confs[:, 1], confs[:, 0])
                                        insert_index = (default_widths > grouped_widths)[lambda x: x.isin([True])].index
                                        conf_index = (default_widths.reset_index(drop=True) >
                                                      grouped_widths)[lambda x: x.isin([True])].index

                                        result.loc[insert_index, 'lower'] = confs[conf_index, 0]
                                        result.loc[insert_index, 'upper'] = confs[conf_index, 1]
                                        result.loc[insert_index, 'significance'] = significances[conf_index]

                                else:
                                    all_ranges = np.array([icp.predict(X.values)])
                                    all_confs = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)
                                    significances = get_categorical_conf(all_confs)
                                    result.loc[X.index, 'significance'] = significances.flatten()

                row_insights['confidence'] = result['significance']

                if is_numerical:
                    row_insights['lower'] = result['lower']
                    row_insights['upper'] = result['upper']

                # anomaly detection
                if is_anomaly_task:
                    row_insights['anomaly'] = None
                    if ns.target_name in ns.data.columns:
                        anomalies = get_anomalies(row_insights,
                                                  ns.data[ns.target_name],
                                                  cooldown=ns.pred_args.anomaly_cooldown)
                        if '__mdb_ts_inferred' in ns.data.columns:
                            ignore_idxs = np.where(ns.data['__mdb_ts_inferred'].values)
                            anomalies[ignore_idxs] = None
                        row_insights['anomaly'] = list(anomalies)

            if ns.tss.is_timeseries and ns.tss.horizon > 1:
                if is_numerical and ns.pred_args.simple_ts_bounds:
                    row_insights = add_tn_num_conf_bounds(row_insights, ns.tss)
                elif not is_numerical:
                    row_insights = add_tn_cat_conf_bounds(row_insights, ns.tss)

            row_insights, global_insights = self._formatted(row_insights, global_insights, ns, is_numerical)

        return row_insights, global_insights

    @staticmethod
    def _formatted(row_insights, global_insights, ns, is_numerical):
        # clip if necessary
        if is_numerical:
            lower_limit = 0.0 if ns.positive_domain else -pow(2, 62)
            upper_limit = pow(2, 62)
            if not (ns.tss.is_timeseries and ns.tss.horizon > 1):
                row_insights['upper'] = row_insights['upper'].clip(lower_limit, upper_limit)
                row_insights['lower'] = row_insights['lower'].clip(lower_limit, upper_limit)
                row_insights['prediction'] = row_insights['prediction'].clip(lower_limit, upper_limit)
            else:
                row_insights['upper'] = [np.array(row).clip(lower_limit, upper_limit).tolist()
                                         for row in row_insights['upper']]
                row_insights['lower'] = [np.array(row).clip(lower_limit, upper_limit).tolist()
                                         for row in row_insights['lower']]
                row_insights['prediction'] = [np.array(row).clip(lower_limit, upper_limit).tolist()
                                              for row in row_insights['prediction']]

        # Make sure the target and real values are of an appropriate type
        if ns.target_dtype in (dtype.integer, ):
            row_insights['prediction'] = row_insights['prediction'].astype(int)
            row_insights['upper'] = row_insights['upper'].astype(int)
            row_insights['lower'] = row_insights['lower'].astype(int)

        elif ns.target_dtype in (dtype.float, dtype.quantity):
            row_insights['prediction'] = row_insights['prediction'].astype(float)
            row_insights['upper'] = row_insights['upper'].astype(float)
            row_insights['lower'] = row_insights['lower'].astype(float)

        elif ns.target_dtype in (dtype.short_text, dtype.rich_text, dtype.binary, dtype.categorical):
            row_insights['prediction'] = row_insights['prediction'].astype(str)

        # horizon collapse
        if ns.tss.is_timeseries and is_numerical and ns.tss.horizon > 1:
            row_insights['prediction_sum'] = row_insights['prediction'].apply(lambda x: sum(x))
            row_insights['lower_sum'] = row_insights['lower'].apply(lambda x: min(x))
            row_insights['upper_sum'] = row_insights['upper'].apply(lambda x: max(x))
            row_insights['confidence_mean'] = row_insights['confidence'].apply(lambda x: np.mean(x))

        return row_insights, global_insights

    @staticmethod
    def _ts_assign_confs(result, df, confs, significances, tss) -> pd.DataFrame:
        result.loc[df.index, 'lower'] = confs[:, 0, 0]
        result.loc[df.index, 'upper'] = confs[:, 0, 1]
        result.loc[df.index, 'significance'] = significances[:, 0]
        for timestep in range(1, tss.horizon):
            result.loc[df.index, f'lower_timestep_{timestep}'] = confs[:, timestep, 0]
            result.loc[df.index, f'upper_timestep_{timestep}'] = confs[:, timestep, 1]
            result.loc[df.index,
                       f'significance_timestep_{timestep}'] = significances[:, timestep]

        # TODO: only if tighter
        # merge all significances, lower and upper bounds into a single column
        for base_col in ['significance', 'lower', 'upper']:
            added_cols = [f'{base_col}_timestep_{t}' for t in range(1, tss.horizon)]
            cols = [base_col] + added_cols
            result.loc[df.index, base_col] = result.loc[df.index, cols].values.tolist()

        return result



from copy import deepcopy
from typing import Dict, List, Iterable, Tuple

import torch
import numpy as np
from scipy.special import softmax

from lightwood.helpers.log import log
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.constants import _UNCOMMON_WORD


class OneHotEncoder(BaseEncoder):
    """
    Creates a one-hot encoding (OHE) for categorical data. One-hot encoding represents categorical information as a vector where each individual dimension corresponds to a category. A category has a 1:1 mapping between dimension indicated by a "1" in that position. For example, imagine 3 categories, :math:`A`, :math:`B`, and :math:`C`; these can be represented as follows:

    .. math::

       A &= [1, 0, 0] \\
       B &= [0, 1, 0] \\
       C &= [0, 0, 1]
    
    The OHE encoder operates in 2 modes:
        (1) "use_unknown=True": Makes an :math:`N+1` length vector for :math:`N` categories, the first index always corresponds to the unknown category.

        (2) "use_unknown=False": Makes an :math:`N` length vector for :math:`N` categories, where an empty vector of 0s indicates an unknown/missing category.

    An encoder can represent a feature column or target column; in this case it represents a target, `is_target` is `True`, and `target_weights`. The `target_weights` parameter enables users to specify how heavily each class should be weighted within a mixer - useful in imbalanced classes.

    By default, `dataprep_ml.StatisticalAnalysis` phase will provide `target_weights` as the relative fraction of each class in the data which is important for imbalanced populations; for example, suppose there is a 80/05/15 imbalanced representation across 3 different classes - `target_weights` will be a vector as such::

    target_weights = {"class1": 0.8, "class2": 0.05, "class3": 0.15}

    Users should note that models will be presented with the inverse of the target weights, `inv_target_weights`, which will perform the 1/target_value_per_class operation. **This means large values will result in small weights for the model**.
    """ # noqa
    def __init__(
        self,
        is_target: bool = False,
        target_weights: Dict[str, float] = None,
        use_unknown: bool = True,
    ):
        """
        :param is_target: True if this encoder featurizes the target column
        :param target_weights: Percentage of total population represented by each category (between [0, 1]).
        :param mode: True uses an extra dimension to account for unknown/out-of-distribution categories
        """  # noqa

        super().__init__(is_target)
        self.map = None  # category name -> index
        self.rev_map = None  # index -> category name
        self.use_unknown = use_unknown

        # Weight-balance info if encoder represents target
        self.target_weights = None
        self.index_weights = None  # vector-weights, mapped by class id
        if self.is_target:
            self.target_weights = deepcopy(target_weights)

    def prepare(self, priming_data: Iterable[str]):
        """
        Prepares the OHE Encoder by creating a dictionary mapping.

        Unknown categories must be explicitly handled as python `None` types.
        """ # noqa
        if self.is_prepared:
            raise Exception('You can only call "prepare" once for a given encoder.')

        unq_cats = np.unique([i for i in priming_data if i is not None]).tolist()

        if self.use_unknown:
            log.debug("Encoding UNKNOWN categories as index 0")
            self.map = {cat: indx + 1 for indx, cat in enumerate(unq_cats)}
            self.map.update({_UNCOMMON_WORD: 0})
            self.rev_map = {indx: cat for cat, indx in self.map.items()}
        else:
            log.debug("Encoding UNKNOWN categories as vector of all 0s")
            self.map = {cat: indx for indx, cat in enumerate(unq_cats)}
            self.rev_map = {indx: cat for cat, indx in self.map.items()}

        # Set the length of output
        self.output_size = len(self.map)

        # For target-only, report on relative weights of classes
        # Each dimension of the inv_target_weights respects `map`
        if self.is_target:

            # Equally wt. all classes
            self.index_weights = torch.ones(size=(self.output_size,))

            # If imbalanced detected, weight by inverse
            if self.target_weights is not None:

                # Check target weights properly specified
                if sum([np.isclose(i, 0) for i in self.target_weights.values()]) > 0:
                    raise ValueError('Target weights cannot be 0')

                for cat in self.map.keys():
                    if cat != _UNCOMMON_WORD:
                        self.index_weights[self.map[cat]] = self.target_weights[cat]

                # If using an unknown category, set to smallest possible value
                if self.use_unknown:
                    self.target_weights[_UNCOMMON_WORD] = np.min(list(self.target_weights.values()))
                    self.index_weights[0] = self.target_weights[_UNCOMMON_WORD]

        self.is_prepared = True

    def encode(self, column_data: Iterable[str]) -> torch.Tensor:
        """
        Encodes pre-processed data into OHE. Unknown/unrecognized classes vector of all 0s.

        :param column_data: Pre-processed data to encode
        :returns: Encoded data of form :math:`N_{rows} x N_{categories}`
        """ # noqa
        if not self.is_prepared:
            raise Exception(
                'You need to call "prepare" before calling "encode" or "decode".'
            )

        ret = torch.zeros(size=(len(column_data), self.output_size))

        for idx, word in enumerate(column_data):
            index = self.map.get(word, None)

            if index is not None:
                ret[idx, index] = 1

            if self.use_unknown and index is None:
                ret[idx, 0] = 1

        return torch.Tensor(ret)

    def decode(self, encoded_data: torch.Tensor):
        """
        Decodes OHE mapping into the original categories. Since this approach uses an argmax, decoding flexibly works either on logits or an explicitly OHE vector.

        :param: encoded_data:
        :returns Returns the original category names for encoded data.
        """ # noqa
        encoded_data_list = encoded_data.tolist()
        ret = []

        for vector in encoded_data_list:

            all_zeros = not np.any(vector)
            if not self.use_unknown and all_zeros:
                ret.append(_UNCOMMON_WORD)
            else:
                ret.append(self.rev_map[np.argmax(vector)])

        return ret

    def decode_probabilities(self, encoded_data: torch.Tensor) -> Tuple[List[str], List[List[float]], Dict[int, str]]:
        """
        Provides decoded answers, as well as a probability assignment to each data point.

        :param encoded_data: the output of a mixer model

        :returns Decoded values for each data point, Probability vector for each category, and the reverse map of dimension to category name
        """ # noqa
        encoded_data_list = encoded_data.tolist()
        ret = []
        probs = []

        for vector in encoded_data_list:
            if not np.any(vector):  # Vector of all 0s -> unknown category
                ret.append(_UNCOMMON_WORD)
            else:
                ret.append(self.rev_map[np.argmax(vector)])

            probs.append(self._norm_vec(vector))

        return ret, probs, self.rev_map

    @staticmethod
    def _norm_vec(vec: List[float]):
        """
        Given a vector, normalizes so that the sum of elements is 1, using softmax.

        :param vec: Assigned weights for each category
        """
        return softmax(vec)



import torch
import numpy as np
from lightwood.encoder import BaseEncoder
from sklearn.preprocessing import MultiLabelBinarizer


class MultiHotEncoder(BaseEncoder):
    def __init__(self, is_target: bool = False):
        super().__init__(is_target)
        self._binarizer = MultiLabelBinarizer()
        self._seen = set()
        self.output_size = None

    @staticmethod
    def _clean_col_data(column_data):
        column_data = [(arr if arr is not None else []) for arr in column_data]
        column_data = [[str(x) for x in arr] for arr in column_data]
        return column_data

    def prepare(self, priming_data, max_dimensions=100):
        priming_data = self._clean_col_data(priming_data)
        self._binarizer.fit(priming_data + [('None')])
        for arr in priming_data:
            for x in arr:
                self._seen.add(x)
        self.is_prepared = True
        self.output_size = len(self.encode(priming_data[0:1])[0])

    def encode(self, column_data):
        column_data = self._clean_col_data(column_data)
        data_array = self._binarizer.transform(column_data)
        return torch.Tensor(data_array)

    def decode(self, vectors):
        # It these are logits output by the neural network, we need to treshold them to binary vectors
        vectors = np.where(vectors > 0, 1, 0)
        words_tuples = self._binarizer.inverse_transform(vectors)
        return [list(w) for w in words_tuples]



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
import os

import nltk
from contextlib import redirect_stdout

from lightwood.helpers.log import log


try:
    with redirect_stdout(open(os.devnull, "w")):
        nltk.data.find('tokenizers/punkt')
except LookupError:
    try:
        nltk.download('punkt', quiet=True)
    except Exception:
        log.error("NLTK was unable to download the 'punkt' package. Please check your connection and try again!")

try:
    with redirect_stdout(open(os.devnull, "w")):
        from nltk.corpus import stopwords
        stopwords.words('english')
except LookupError:
    try:
        nltk.download('stopwords', quiet=True)
    except Exception:
        log.error("NLTK was unable to download the 'stopwords' package. Please check your connection and try again!")



from typing import Optional, List, Dict
import torch
import pandas as pd

from dataprep_ml import StatisticalAnalysis

from lightwood.helpers.log import log, timed
from lightwood.api.types import ProblemDefinition, PredictionArguments
from lightwood.helpers.ts import get_inferred_timestamps
from lightwood.analysis.base import BaseAnalysisBlock


@timed
def explain(data: pd.DataFrame,
            encoded_data: torch.Tensor,
            predictions: pd.DataFrame,
            target_name: str,
            target_dtype: str,

            problem_definition: ProblemDefinition,
            stat_analysis: StatisticalAnalysis,
            pred_args: PredictionArguments,
            runtime_analysis: Dict,

            explainer_blocks: Optional[List[BaseAnalysisBlock]] = [],
            ts_analysis: Optional[Dict] = {}
            ):
    """
    This procedure runs at the end of every normal `.predict()` call. Its goal is to generate prediction insights,
    potentially using information generated at the model analysis stage (e.g. confidence estimation).

    As in `analysis()`, any user-specified analysis blocks (see class `BaseAnalysisBlock`) are also called here.

    :return:
    row_insights: a DataFrame containing predictions and all generated insights at a row-level.
    """

    # ------------------------- #
    # Setup base insights
    # ------------------------- #
    predictions = predictions.reset_index(drop=True)
    data = data.reset_index(drop=True)
    tss = problem_definition.timeseries_settings
    row_insights = pd.DataFrame()
    global_insights = {}

    def _reformat_ts_columns(tss, out_df, in_df):
        if tss.is_timeseries:
            if tss.group_by:
                for col in tss.group_by:
                    out_df[f'group_{col}'] = in_df[col]

            out_df[f'order_{tss.order_by}'] = in_df[tss.order_by]
            out_df[f'order_{tss.order_by}'] = get_inferred_timestamps(
                out_df, tss.order_by, ts_analysis['deltas'], tss, stat_analysis,
                time_format=pred_args.time_format
            )
        return out_df

    if not explainer_blocks:
        predictions.rename(columns={'__mdb_original_index': 'original_index'}, inplace=True)
        predictions = _reformat_ts_columns(tss, predictions, data)
        return predictions, global_insights

    row_insights['original_index'] = data['__mdb_original_index']
    row_insights['prediction'] = predictions['prediction']

    if pred_args.predict_proba:
        for col in predictions.columns:
            if '__mdb_proba' in col:
                row_insights[col] = predictions[col]

    row_insights = _reformat_ts_columns(tss, row_insights, data)

    kwargs = {
        'data': data,
        'encoded_data': encoded_data,
        'predictions': predictions,
        'analysis': runtime_analysis,
        'target_name': target_name,
        'target_dtype': target_dtype,
        'tss': tss,
        'positive_domain': stat_analysis.positive_domain,
        'anomaly_detection': problem_definition.anomaly_detection,
        'pred_args': pred_args
    }

    # ------------------------- #
    # Call explanation blocks
    # ------------------------- #
    for block in explainer_blocks:
        log.info("The block %s is now running its explain() method", block.__class__.__name__)
        row_insights, global_insights = block.explain(row_insights, global_insights, **kwargs)

    return row_insights, global_insights



import dill
from typing import Dict, Optional

import pandas as pd
from lightwood.api.types import ModelAnalysis


# Interface that must be respected by predictor objects generated from JSON ML and/or compatible with Mindsdb
class PredictorInterface:
    """
    Abstraction of a Lightwood predictor. The ``PredictorInterface`` encompasses how Lightwood interacts with the full ML pipeline. Internally,

    The ``PredictorInterface`` class must have several expected functions:

    - ``analyze_data``: Peform a statistical analysis on the unprocessed data; this helps inform downstream encoders and mixers on how to treat the data types.
    - ``preprocess``: Apply cleaning functions to each of the columns within the dataset to prepare them for featurization
    - ``split``: Split the input dataset into a train/dev/test set according to your splitter function
    - ``prepare``: Create and, if necessary, train your encoders to create feature representations from each column of your data.
    - ``featurize``: For input, pre-processed data, create feature vectors
    - ``fit``: Train your mixer models to yield predictions from featurized data
    - ``analyze_ensemble``: Evaluate the quality of fit for your mixer models
    - ``adjust``: Incorporate new data to update pre-existing model(s).

    For simplification, we offer an end-to-end approach that allows you to input raw data and follow every step of the process until you reach a trained predictor with the ``learn`` function:

        - ``learn``: An end-to-end technique specifying how to pre-process, featurize, and train the model(s) of interest. The expected input is raw, untrained data. No explicit output is provided, but the Predictor object will "host" the trained model thus.

    You can also use the predictor to now estimate new data:

    - ``predict``: Deploys the chosen best model, and evaluates the given data to provide target estimates.
    - ``test``: Similar to predict, but user also passes an accuracy function that will be used to compute a metric with the generated predictions.
    - ``save``: Saves the Predictor object for further use.

    The ``PredictorInterface`` is created via J{ai}son's custom code creation. A problem inherits from this class with pre-populated routines to fill out expected results, given the nature of each problem type.
    """ # noqa

    model_analysis: ModelAnalysis = None

    def __init__(self):
        pass

    def analyze_data(self, data: pd.DataFrame) -> None:
        """
        Performs a statistical analysis on the data to identify distributions, imbalanced classes, and other nuances within the data.

        :param data: Data used in training the model(s).
        """ # noqa
        pass

    def preprocess(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Cleans the unprocessed dataset provided.

        :param data: (Unprocessed) Data used in training the model(s).
        :returns: The cleaned data frame
        """ # noqa
        pass

    def split(self, data: pd.DataFrame) -> Dict[str, pd.DataFrame]:
        """
        Categorizes the data into a training/testing split; if data is a classification problem, will stratify the data.

        :param data: Pre-processed data, but generically any dataset to split into train/dev/test.
        :returns: Dictionary containing training/testing fraction
        """ # noqa
        pass

    def prepare(self, data: Dict[str, pd.DataFrame]) -> None:
        """
        Prepares the encoders for each column of data.

        :param data: Pre-processed data that has been split into train/test. Explicitly uses "train" and/or "dev" in preparation of encoders.

        :returns: Nothing; prepares the encoders for learned representations.
        """  # noqa

    def featurize(self, split_data: Dict[str, pd.DataFrame]):
        """
        Provides an encoded representation for each dataset in ``split_data``. Requires `self.encoders` to be prepared.

        :param split_data: Pre-processed data from the dataset, split into train/test (or any other keys relevant)

        :returns: For each dataset provided in ``split_data``, the encoded representations of the data.
        """ # noqa
        pass

    def fit(self, enc_data: Dict[str, pd.DataFrame]) -> None:
        """
        Fits "mixer" models to train predictors on the featurized data. Instantiates a set of trained mixers and an ensemble of them.

        :param enc_data: Pre-processed and featurized data, split into the relevant train/test splits. Keys expected are "train", "dev", and "test"
        """  # noqa
        pass

    def analyze_ensemble(self, enc_data: Dict[str, pd.DataFrame]) -> None:
        """
        Evaluate the quality of mixers within an ensemble of models.

        :param enc_data: Pre-processed and featurized data, split into the relevant train/test splits.
        """
        pass

    def learn(self, data: pd.DataFrame) -> None:
        """
        Trains the attribute model starting from raw data. Raw data is pre-processed and cleaned accordingly. As data is assigned a particular type (ex: numerical, categorical, etc.), the respective feature encoder will convert it into a representation useable for training ML models. Of all ML models requested, these models are compiled and fit on the training data.

        This step amalgates ``preprocess`` -> ``featurize`` -> ``fit`` with the necessary splitting + analyze_data that occurs.

        :param data: (Unprocessed) Data used in training the model(s).

        :returns: Nothing; instantiates with best fit model from ensemble.
        """  # noqa
        pass

    def adjust(self, new_data: pd.DataFrame, old_data: Optional[pd.DataFrame] = None, adjust_args: Optional[dict] = None
               ) -> None:
        """
        Adjusts a previously trained model on new data. Adopts the same process as ``learn`` but with the exception that the `adjust` function expects the best model to have been already trained.

        .. warning:: This is experimental and subject to change.
        :param new_data: New data used to adjust a previously trained model.
        :param old_data: In some situations, the old data is still required to train a model (i.e. Regression mixer) to ensure the new data doesn't entirely override it.
        :param adjust_args: Optional dictionary with parameters to customize the finetuning process.

        :returns: Adjusts best-fit model in-place, doesn't return anything.
        """  # noqa
        pass

    def predict(self, data: pd.DataFrame, args: Dict[str, object] = {}) -> pd.DataFrame:
        """
        Intakes raw data to provide model predictions.

        :param data: Data (n_samples, n_columns) that the model will use as input to predict the corresponding target value for each sample.
        :param args: any parameters used to customize inference behavior. Wrapped as a ``PredictionArguments`` object.

        :returns: A dataframe containing predictions and additional sample-wise information. `n_samples` rows.
        """  # noqa
        pass

    def test(
            self, data: pd.DataFrame, metrics: list, args: Dict[str, object] = {}, strict: bool = False
    ) -> pd.DataFrame:
        """
        Intakes raw data to compute values for a list of provided metrics using a Lightwood predictor.

        :param data: Data (n_samples, n_columns) that the model(s) will evaluate on and provide the target prediction.
        :param metrics: A list of metrics to evaluate the model's performance on.
        :param args: parameters needed to update the predictor ``PredictionArguments`` object, which holds any parameters relevant for prediction.
        :param strict: If True, the function will raise an error if the model does not support any of the requested metrics. Otherwise it skips them.

        :returns: A dataframe with `n_metrics` columns, each cell containing the respective score of each metric.
        """  # noqa
        pass

    def save(self, file_path: str) -> None:
        """
        With a provided file path, saves the Predictor instance for later use.
        :param file_path: Location to store your Predictor Instance.
        :returns: Saves Predictor instance.
        """
        with open(file_path, "wb") as fp:
            dill.dump(self, fp)

    def export(self, file_path: str, json_ai_code: str) -> None:
        """
        Exports both the predictor object and its code to a single binary file for later usage.

        :param file_path: Location to store your Predictor Instance.
        :param json_ai_code: The code generated by the user's specification.

        :returns: Saves Predictor instance.
        """

        predictor_dict = {}
        predictor_dict['predictor'] = self
        predictor_dict['code'] = json_ai_code

        with open(file_path, "wb") as fp:
            dill.dump(predictor_dict, fp)



from typing import Optional
import torch
import pandas as pd
from scipy.special import softmax
from sklearn.linear_model import Ridge

from lightwood.helpers.log import log
from type_infer.dtype import dtype
from lightwood.mixer import BaseMixer
from lightwood.encoder.base import BaseEncoder
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import ConcatedEncodedDs, EncodedDs


class Regression(BaseMixer):
    """
    The `Regression` mixer inherits from scikit-learn's `Ridge` class
    (https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html)
    
    This class performs Ordinary Least-squares Regression (OLS) under the hood; 
    this means it fits a set of coefficients (w_1, w_2, ..., w_N) for an N-length feature vector, that minimize the difference
    between the predicted target value and the observed true value.
  
    This mixer intakes featurized (encoded) data to predict the target. It deploys if the target data-type is considered numerical/integer.
    """ # noqa
    model: Ridge
    label_map: dict
    supports_proba: bool

    def __init__(self, stop_after: float, target_encoder: BaseEncoder, dtype_dict: dict, target: str):
        """
        :param stop_after: Maximum amount of seconds it should fit for, currently ignored
        :param target_encoder: The encoder which will be used to decode the target
        :param dtype_dict: A map of feature names and their data types
        :param target: Name of the target column
        """ # noqa
        super().__init__(stop_after)
        self.target_encoder = target_encoder
        self.target_dtype = dtype_dict[target]
        self.dtype_dict = dtype_dict
        self.supports_proba = self.target_dtype in [dtype.binary, dtype.categorical]
        self.label_map = {}
        self.stable = False

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits `Ridge` model on input feature data to provide predictions.

        :param train_data: Regression if fit on this
        :param dev_data: This just gets concatenated to the ``train_data``
        """
        if self.target_dtype not in (dtype.float, dtype.integer, dtype.quantity):
            raise Exception(f'Unspported {self.target_dtype} type for regression')

        if self.stop_after < len(train_data) * len(self.dtype_dict) / pow(10, 3):
            raise Exception(f'Insufficient time ({self.stop_after} seconds) to fit a linear regression on the data!')

        log.info('Fitting Linear Regression model')
        X = []
        Y = []
        for x, y in ConcatedEncodedDs([train_data, dev_data]):
            X.append(x.tolist())
            Y.append(y.tolist())

        if self.supports_proba:
            self.label_map = self.target_encoder.rev_map

        self.model = Ridge().fit(X, Y)
        log.info(f'Regression based correlation of: {self.model.score(X, Y)}')

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs, args: Optional[dict] = None) -> None:
        """
        Fits the linear regression on some data, this refits the model entirely rather than updating it

        :param train_data: Regression is fit on this
        :param dev_data: This just gets concatenated to the ``train_data``
        """
        self.fit(train_data, dev_data)

    def __call__(self, ds: EncodedDs,
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Make predictions based on datasource with the same features as the ones used for fitting

        :param ds: Predictions are generate from it
        :param arg: Any additional arguments used in predicting

        :returns: A dataframe cotaining the decoded predictions and (depending on the args) additional information such as the probabilites for each target class
        """ # noqa
        X = ds.get_encoded_data(include_target=False)
        Yh = self.model.predict(X)

        decoded_predictions = self.target_encoder.decode(torch.Tensor(Yh))

        ydf = pd.DataFrame({'prediction': decoded_predictions})

        if args.predict_proba and self.label_map:
            raw_predictions = softmax(Yh.squeeze(), axis=1)
            for idx, label in enumerate(self.target_encoder.rev_map.values()):
                ydf[f'__mdb_proba_{label}'] = raw_predictions[:, idx]

        return ydf



import os
from typing import Dict
import psutil
import multiprocessing as mp
from lightwood.helpers.log import log

MAX_SEQ_ENCODERS = 20
MAX_SEQ_LEN = 100_000


def get_nr_procs(df=None):
    if 'LIGHTWOOD_N_WORKERS' in os.environ:
        try:
            n = int(os.environ['LIGHTWOOD_N_WORKERS'])
        except ValueError:
            n = 1
        return n
    elif os.name == 'nt':
        return 1
    else:
        available_mem = psutil.virtual_memory().available
        max_per_proc_usage = 2 * pow(10, 8)

        if df is not None:
            max_per_proc_usage += df.memory_usage(index=True, deep=True).sum()
        proc_count = min(mp.cpu_count(), available_mem // max_per_proc_usage) - 1

        return max(proc_count, 1)


def run_mut_method(obj: object, arg: object, method: str, identifier: str) -> str:
    try:
        obj.__getattribute__(method)(arg)
        return obj, identifier
    except Exception as e:
        log.error(f'Exception {e} when running with identifier {identifier}')
        raise e


def mut_method_call(object_dict: Dict[str, tuple]) -> Dict[str, object]:
    manager = mp.Manager()
    return_dict = manager.dict()

    nr_procs = get_nr_procs()
    pool = mp.Pool(processes=nr_procs)
    promise_arr = []
    for name, data in object_dict.items():
        promise = pool.apply_async(func=run_mut_method, args=(data[0], data[1], data[2], name))
        promise_arr.append(promise)

    for promise in promise_arr:
        obj, identifier = promise.get()
        return_dict[identifier] = obj
        log.info(f'Done running for: {identifier}')

    pool.close()
    pool.join()

    return dict(return_dict)


def parallel_encoding_check(df, encoders):
    """
      Given a dataframe and some encoders, this rule-based method determines whether to train these encoders in parallel.
      This has runtime implications, as instancing a new Lightwood process has noticeable overhead.
    """  # noqa
    trainable_encoders = [enc for col, enc in encoders.items() if enc.is_trainable_encoder]

    if len(trainable_encoders) > MAX_SEQ_ENCODERS:
        return True

    if len(df) > MAX_SEQ_LEN:
        return True

    return False



import os
import time
from typing import Iterable
from collections import deque

import numpy as np
import torch
from torch.utils.data import DataLoader
import pandas as pd
from type_infer.dtype import dtype
from transformers import (
    DistilBertModel,
    DistilBertForSequenceClassification,
    DistilBertTokenizerFast,
    AdamW,
    get_linear_schedule_with_warmup,
)
from sklearn.model_selection import train_test_split

from lightwood.encoder.text.helpers.pretrained_helpers import TextEmbed
from lightwood.helpers.device import get_device_from_name
from lightwood.encoder.base import BaseEncoder
from lightwood.helpers.log import log
from lightwood.helpers.torch import LightwoodAutocast
from lightwood.helpers.general import is_none


class PretrainedLangEncoder(BaseEncoder):
    is_trainable_encoder: bool = True

    """
    Creates a contextualized embedding to represent input text via the [CLS] token vector from DistilBERT (transformers). (Sanh et al. 2019 - https://arxiv.org/abs/1910.01108).

    In certain text tasks, this model can use a transformer to automatically fine-tune on a class of interest (providing there is a 2 column dataset, where the input column is text).

    """ # noqa

    def __init__(
        self,
        stop_after: float,
        is_target: bool = False,
        batch_size: int = 10,
        max_position_embeddings: int = None,
        frozen: bool = False,
        epochs: int = 1,
        output_type: str = None,
        embed_mode: bool = True,
        device: str = '',
    ):
        """
        :param is_target: Whether this encoder represents the target. NOT functional for text generation yet.
        :param batch_size: size of batch while fine-tuning
        :param max_position_embeddings: max sequence length of input text
        :param frozen: If True, freezes transformer layers during training.
        :param epochs: number of epochs to train model with
        :param output_type: Data dtype of the target; if categorical/binary, the option to return logits is possible.
        :param embed_mode: If True, assumes the output of the encode() step is the CLS embedding (this can be trained or not). If False, returns the logits of the tuned task.
        :param device: name of the device that get_device_from_name will attempt to use.
        """ # noqa
        super().__init__(is_target)

        self.output_type = output_type
        self.name = "distilbert text encoder"

        self._max_len = max_position_embeddings
        self._frozen = frozen
        self._batch_size = batch_size
        self._epochs = epochs
        self._patience = 3  # measured in batches rather than epochs
        self._val_loss_every = 5  # how many batches to wait before checking val loss. If -1, will check train loss instead of val for early stopping.  # noqa
        self._tr_loss_every = 2  # same as above, but only applies if `_val_loss_every` is set to -1

        # Model setup
        self._model = None
        self.model_type = None

        self._classifier_model_class = DistilBertForSequenceClassification
        self._embeddings_model_class = DistilBertModel
        self._pretrained_model_name = "distilbert-base-uncased"
        self._tokenizer = DistilBertTokenizerFast.from_pretrained(self._pretrained_model_name)

        self.device = get_device_from_name(device)

        self.stop_after = stop_after

        self.embed_mode = embed_mode
        self.uses_target = True
        self.output_size = None

        if self.embed_mode:
            log.info("Embedding mode on. [CLS] embedding dim output of encode()")
        else:
            log.info("Embedding mode off. Logits are output of encode()")

    def prepare(
        self,
        train_priming_data: pd.Series,
        dev_priming_data: pd.Series,
        encoded_target_values: torch.Tensor,
    ):
        """
        Fine-tunes a transformer on the priming data.

        Transformer is fine-tuned with weight-decay on training split. 
        
        Train + Dev are concatenated together and a transformer is then fine tuned with weight-decay applied on the transformer parameters. The option to freeze the underlying transformer and only train a linear layer exists if `frozen=True`. This trains faster, with the exception that the performance is often lower than fine-tuning on internal benchmarks.
        
        :param train_priming_data: Text data in the train set
        :param dev_priming_data: Text data in the dev set
        :param encoded_target_values: Encoded target labels in Nrows x N_output_dimension
        """ # noqa
        if self.is_prepared:
            raise Exception("Encoder is already prepared.")

        os.environ['TOKENIZERS_PARALLELISM'] = 'true'

        # remove empty strings (`None`s for dtype `object`)
        filtered_tr = train_priming_data[~train_priming_data.isna()]
        filtered_dev = dev_priming_data[~dev_priming_data.isna()]

        if filtered_dev.shape[0] > 0:
            priming_data = pd.concat([filtered_tr, filtered_dev]).tolist()
            val_size = (len(dev_priming_data)) / len(train_priming_data)
        else:
            priming_data = filtered_tr.tolist()
            val_size = 0.1  # leave out 0.1 for validation

        # Label encode the OHE/binary output for classification
        labels = encoded_target_values.argmax(dim=1)

        # Split into train and validation sets
        train_texts, val_texts, train_labels, val_labels = train_test_split(priming_data, labels, test_size=val_size)

        # If classification, then fine-tune
        if self.output_type in (dtype.categorical, dtype.binary):
            log.info("Training model.\n\tOutput trained is categorical")

            # Prepare priming data into tokenized form + attention masks
            training_text = self._tokenizer(train_texts, truncation=True, padding=True)
            validation_text = self._tokenizer(val_texts, truncation=True, padding=True)

            # Construct the model
            self._model = self._classifier_model_class.from_pretrained(
                self._pretrained_model_name,
                num_labels=len(encoded_target_values[0]),  # max classes to test
            ).to(self.device)

            # Construct the dataset for training
            xinp = TextEmbed(training_text, train_labels)
            train_dataset = DataLoader(xinp, batch_size=self._batch_size, shuffle=True)

            # Construct the dataset for validation
            xvalinp = TextEmbed(validation_text, val_labels)
            val_dataset = DataLoader(xvalinp, batch_size=self._batch_size, shuffle=True)

            # Set max length of input string; affects input to the model
            if self._max_len is None:
                self._max_len = self._model.config.max_position_embeddings

            if self._frozen:
                log.info("\tFrozen Model + Training Classifier Layers")
                """
                Freeze the base transformer model and train a linear layer on top
                """
                # Freeze all the transformer parameters
                for param in self._model.base_model.parameters():
                    param.requires_grad = False

                optimizer_grouped_parameters = self._model.parameters()

            else:
                log.info("\tFine-tuning model")
                """
                Fine-tuning parameters with weight decay
                """
                no_decay = [
                    "bias",
                    "LayerNorm.weight",
                ]  # decay on all terms EXCLUDING bias/layernorms
                optimizer_grouped_parameters = [
                    {
                        "params": [
                            p
                            for n, p in self._model.named_parameters()
                            if not any(nd in n for nd in no_decay)
                        ],
                        "weight_decay": 0.01,
                    },
                    {
                        "params": [
                            p
                            for n, p in self._model.named_parameters()
                            if any(nd in n for nd in no_decay)
                        ],
                        "weight_decay": 0.0,
                    },
                ]

            optimizer = AdamW(optimizer_grouped_parameters, lr=1e-5)
            scheduler = get_linear_schedule_with_warmup(
                optimizer,
                num_warmup_steps=0,  # default value for GLUE
                num_training_steps=len(train_dataset) * self._epochs,
            )

            # Train model; declare optimizer earlier if desired.
            self._tune_model(
                train_dataset, val_dataset, optim=optimizer, scheduler=scheduler, n_epochs=self._epochs
            )

        else:
            log.info("Target is not classification; Embeddings Generator only")

            self.model_type = "embeddings_generator"
            self._model = self._embeddings_model_class.from_pretrained(
                self._pretrained_model_name
            ).to(self.device)

            # TODO: Not a great flag
            # Currently, if the task is not classification, you must have an embedding generator only
            if self.embed_mode is False:
                log.info("Embedding mode must be ON for non-classification targets.")
                self.embed_mode = True

        self.is_prepared = True
        encoded = self.encode(priming_data[0:1])
        self.output_size = len(encoded[0])

    def _tune_model(self, train_dataset, val_dataset, optim, scheduler, n_epochs=1):
        """
        Given a model, tune for n_epochs.

        train_dataset - torch.DataLoader; dataset to train
        val_dataset - torch.DataLoader; dataset used to compute validation loss + early stopping
        optim - transformers.optimization.AdamW; optimizer
        scheduler - scheduling params
        n_epochs - max number of epochs to train for, provided there is no early stopping

        """ # noqa
        self._model.train()

        if optim is None:
            log.info("No opt. provided, setting all params with AdamW.")
            optim = AdamW(self._model.parameters(), lr=5e-5)
        else:
            log.info("Optimizer provided")

        if scheduler is None:
            log.info("No scheduler provided.")
        else:
            log.info("Scheduler provided.")

        best_tr_loss = best_val_loss = float("inf")
        tr_loss_queue = deque(maxlen=self._patience)
        patience_counter = self._patience

        started = time.time()
        for epoch in range(n_epochs):
            total_loss = 0

            for bidx, batch in enumerate(train_dataset):
                optim.zero_grad()

                with LightwoodAutocast():
                    loss = self._call(batch)

                tr_loss_queue.append(loss.item())
                total_loss += loss.item()

                loss.backward()
                optim.step()
                if scheduler is not None:
                    scheduler.step()
                if time.time() - started > self.stop_after:
                    break

                # val-based early stopping
                if False and (self._val_loss_every != -1) and (bidx % self._val_loss_every == 0):
                    self._model.eval()
                    val_loss = 0

                    for vbatch in val_dataset:
                        val_loss += self._call(vbatch).item()

                    log.info(f"Epoch {epoch+1} train batch {bidx+1} - Validation loss: {val_loss/len(val_dataset)}")
                    if val_loss / len(val_dataset) >= best_val_loss:
                        break

                    best_val_loss = val_loss / len(val_dataset)
                    self._model.train()

                # train-based early stopping
                elif False and (bidx + 1) % self._tr_loss_every == 0:
                    self._model.eval()

                    tr_loss = np.average(tr_loss_queue)
                    log.info(f"Epoch {epoch} train batch {bidx} - Train loss: {tr_loss}")  # noqa
                    self._model.train()

                    if tr_loss >= best_tr_loss and patience_counter == 0:
                        break
                    elif patience_counter > 0:
                        patience_counter -= 1
                    elif tr_loss < best_tr_loss:
                        best_tr_loss = tr_loss
                        patience_counter = self._patience

            if time.time() - started > self.stop_after:
                break
            self._train_callback(epoch, total_loss / len(train_dataset))

    def _call(self, batch):
        inpids = batch["input_ids"].to(self.device)
        attn = batch["attention_mask"].to(self.device)
        labels = batch["labels"].to(self.device)
        outputs = self._model(inpids, attention_mask=attn, labels=labels)
        loss = outputs[0]
        return loss

    def _train_callback(self, epoch, loss):
        log.info(f"{self.name} at epoch {epoch+1} and loss {loss}!")

    def encode(self, column_data: Iterable[str]) -> torch.Tensor:
        """
        Converts each text example in a column into encoded state. This can be either a vector embedding of the [CLS] token (represents the full text input) OR the logits prediction of the output.

        The transformer model is of form:
        transformer base + pre-classifier linear layer + classifier layer

        The embedding returned is of the [CLS] token after the pre-classifier layer; from internal testing, we found the latent space most highly separated across classes. 

        If the encoder represents the logits in classification, returns a soft-maxed output of the class vector.

        :param column_data: List of text data as strings
        :returns: Embedded vector N_rows x Nembed_dim OR logits vector N_rows x N_classes depending on if `embed_mode` is True or not.
        """ # noqa
        if self.is_prepared is False:
            raise Exception("You need to first prepare the encoder.")

        # Set model to testing/eval mode.
        self._model.eval()

        encoded_representation = []

        with torch.no_grad():
            # Set the weights; this is GPT-2
            for text in column_data:

                # Omit NaNs
                if is_none(text):
                    text = ""

                # Tokenize the text with the built-in tokenizer.
                inp = self._tokenizer.encode(
                    text, truncation=True, return_tensors="pt"
                ).to(self.device)

                if self.embed_mode:  # Embedding mode ON; return [CLS]
                    output = self._model.base_model(inp).last_hidden_state[:, 0]

                    # If the model has a pre-classifier layer, use this embedding.
                    if hasattr(self._model, "pre_classifier"):
                        output = self._model.pre_classifier(output)

                else:  # Embedding mode off; return classes
                    output = self._model(inp).logits

                encoded_representation.append(output.detach())

        return torch.stack(encoded_representation).squeeze(1).to('cpu')

    def decode(self, encoded_values_tensor, max_length=100):
        """
        Text generation via decoding is not supported.
        """ # noqa
        raise Exception("Decoder not implemented.")

    def to(self, device, available_devices):
        """
        Converts encoder models to device specified (CPU/GPU)

        Transformers are LARGE models, please run on GPU for fastest implementation.
        """ # noqa
        for v in vars(self):
            attr = getattr(self, v)
            if isinstance(attr, torch.nn.Module):
                attr.to(device)
        return self
